
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Classifier;

import java.util.*;
import Utils.Misc;
import Utils.ValuesDistribution;

/**
* This class is used to represent anticipatory classifiers.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
**/

public class AnticipatoryClassifier extends BasicClassifier 
{
   protected int nbGoodAnticipations = 0;
   protected int nbBadAnticipations = 0;
   protected int lastEvaluation = 0;
   
   protected Anticipation anticipation;
   protected Perception lastPerceptionWhenAnticipatingWell;
   protected Perception lastPerceptionWhenAnticipatingBadly;
   /**
    * ------------- the following part is devoted to reliability exclusively... -------------------
    */
   protected double reliabilitiesArray[];
   protected double reliabilitiesIfGeneralArray[];

    /* --------------------------- end of part --------------------------------------
     */

   /**
    * @param classifier@param init
    * @param initialReliability
    */
   public AnticipatoryClassifier(AnticipatoryClassifier classifier, boolean init, double initialReliability) 
   {
	super((BasicClassifier)classifier);
	if (classifier.anticipation instanceof MACSAnticipation) anticipation = new MACSAnticipation((MACSAnticipation)classifier.anticipation);
	else if (classifier.anticipation instanceof YACSAnticipation) anticipation = new YACSAnticipation((YACSAnticipation)classifier.anticipation);
	else 
	    {
		System.out.println("systemName not known");
		System.exit(-1);
	    }
	reliabilitiesArray = new double[perceptionLength];
	reliabilitiesIfGeneralArray = new double[perceptionLength];
	if (init)
	    {
		lastEvaluation	=	classifier.lastEvaluation;
		for (byte i=0; i<perceptionLength ; i++)
		    {
			reliabilitiesArray[i] = classifier.reliabilitiesArray[i];
			setReliabilitiesIfGeneral(i,classifier.reliabilitiesIfGeneralArray[i]);
		    }

		lastPerceptionWhenAnticipatingWell = classifier.lastPerceptionWhenAnticipatingWell;
		lastPerceptionWhenAnticipatingBadly = classifier.lastPerceptionWhenAnticipatingBadly;
	    }
	else 
	    {
		initReliabilitiesArrays(initialReliability);

		lastPerceptionWhenAnticipatingWell = null;
		lastPerceptionWhenAnticipatingBadly = null;
	    }    
   }
   
   /**
    * @param strValue
    * @param perceptionForm
    * @param actionForm
    * @param params@param systemName
    * @param actionFormString
    * @param initialReliability
    */
   public AnticipatoryClassifier(String systemName, String strValue, String perceptionForm, String actionFormString, double initialReliability) 
   {
	super(perceptionForm,actionFormString);

	byte posSpace1 = (byte)(strValue.indexOf(' '));
	byte posSpace2 = (byte)(strValue.indexOf(' ', posSpace1+1));
	
	condition = new Perception(strValue.substring(0, posSpace1), perceptionForm);
	action = new Action(strValue.substring(posSpace1+1, posSpace2), actionFormString);
	if (systemName.equals("MACS")) anticipation = new MACSAnticipation(strValue.substring(posSpace2+1), perceptionForm);
	else if (systemName.equals("YACS")) anticipation = new YACSAnticipation(strValue.substring(posSpace2+1), perceptionForm);
	else 
	    {
		System.out.println("systemName not known");
		System.exit(-1);
	    }

	perceptionLength = condition.getLength();
	initReliabilitiesArrays(initialReliability);

	lastPerceptionWhenAnticipatingWell = null;
	lastPerceptionWhenAnticipatingBadly = null;    
   }
   
   /**
    * @param perceptionForm
    * @param actionForm
    * @param params@param systemName
    * @param initialReliability
    */
   public AnticipatoryClassifier(String systemName, String perceptionForm, String actionForm, double initialReliability) 
   {
	super(perceptionForm,actionForm);
	if (systemName.equals("MACS")) anticipation = new MACSAnticipation(perceptionForm);
	else if (systemName.equals("YACS")) anticipation = new YACSAnticipation(perceptionForm);
	else 
	    {
		System.out.println("systemName not known");
		System.exit(-1);
	    }
	initReliabilitiesArrays(initialReliability);

	lastPerceptionWhenAnticipatingWell = null;
	lastPerceptionWhenAnticipatingBadly = null;    
   }
   
   /**
    * Returns a clone of the MACSClassifier
    * @return Object
    */
   public Object clone() 
   {
	return new AnticipatoryClassifier(this,true,0.5);    
   }
   
   /**
    * @param p
    * @return int
    */
   public final int compareTo(Object p) 
   {
	int cond = super.compareTo((BasicClassifier)p);
	if (cond !=0) return cond;
	else
	    {
		int anticipa = ((Perception)anticipation).compareTo(((AnticipatoryClassifier)p).getAnticipation());
		if (anticipa !=0) return anticipa;
	    }
	return 0;    
   }
   
   /**
    * @param classifier
    * }@return Classifier.Perception
    */
   public Perception getLastPerceptionWhenAnticipatingWell() 
   {
	return lastPerceptionWhenAnticipatingWell;    
   }
   
   /**
    * @return Classifier.Perception
    */
   public Perception getLastPerceptionWhenAnticipatingBadly() 
   {
	return lastPerceptionWhenAnticipatingBadly;    
   }
   
   /**
    * @return Classifier.Anticipation
    */
   public Anticipation getAnticipation() 
   {
	return anticipation;    
   }
   
   /**
    * @return double
    */
   public double getExplorationFactor() 
   {
	return Misc.boundedLinearFunction(lastEvaluation,10,20);    
   }
   
   /**
    * @param classifier
    * @return boolean true if the two objects are equal, else false
    */
   public boolean equals(Object classifier) 
   {
	return (condition.equals(((AnticipatoryClassifier)classifier).condition) 
		&& action.equals(((AnticipatoryClassifier)classifier).action) 
		&& anticipation.equals(((AnticipatoryClassifier)classifier).anticipation));    
   }
   
   /**
    * @param classifier
    * @param watchForAnticipation
    * @return boolean
    */
   public boolean equals(AnticipatoryClassifier classifier, boolean watchForAnticipation) 
   {
	return
	    (
	     condition.equals(classifier.condition)
	     && action.equals(classifier.action)
	     && ((!watchForAnticipation) || (watchForAnticipation && anticipation.equals(classifier.anticipation)))
	     );    
   }
   
   /**
    * @param classifier
    * @param watchForAnticipation
    * @return boolean
    */
   public boolean isMoreGeneralThan(AnticipatoryClassifier classifier, boolean watchForAnticipation) 
   {
	return
	    (
	     condition.isMoreGeneralThan(classifier.condition)
	     && ( action.equals(classifier.action) || action.isMoreGeneralThan(classifier.action) )
	     && ((!watchForAnticipation) || (watchForAnticipation && 
					     (condition.passThrough(anticipation)).isEqualOrMoreGeneralThan(classifier.condition.passThrough(classifier.anticipation)) ))
	     );    
   }
   
   /**
    * @param classifier
    * @param watchForAnticipation
    * @return boolean
    */
   public boolean isLessGeneralThan(AnticipatoryClassifier classifier, boolean watchForAnticipation) 
   {
	return
	    (
	     condition.isLessGeneralThan(classifier.condition)
	     && ( action.equals(classifier.action) || action.isMoreGeneralThan(classifier.action) )
	     && ((!watchForAnticipation) || (watchForAnticipation && 
					     (condition.passThrough(anticipation)).isEqualOrLessGeneralThan(classifier.condition.passThrough(classifier.anticipation)) ))
	     );    
   }
   
   /**
    * @param classifier
    * @param watchForAnticipation
    * @return boolean
    */
   public boolean matches(AnticipatoryClassifier classifier, boolean watchForAnticipation) 
   {
	return
	    (
	     condition.matches(classifier.condition)
	     && action.matches(classifier.action)
	     && ((!watchForAnticipation) || (watchForAnticipation && 
					     (condition.passThrough(anticipation)).matches(classifier.condition.passThrough(classifier.anticipation))))
	     );    
   }
   
   /**
    * @param ant
    * @return boolean
    */
   public boolean matchAnticipation(Perception ant) 
   {
    	return (anticipation.matches(ant));    
   }
   
   /**
    * @param perception
    * @return Classifier.Perception
    */
   public Perception computeAnticipation(Perception perception) 
   {
	return (perception.passThrough(anticipation));    
   }
   
   /**
    * @param newAnt
    */
   public void setAnticipation(Anticipation newAnt) 
   {
	anticipation = newAnt;    
   }
   
   /**
    * @param nbEvaluationsToSpecialize
    * @return boolean
    */
   public boolean doesOscillate(byte nbEvaluationsToSpecialize) 
   {
	return (((nbGoodAnticipations+nbBadAnticipations)>=nbEvaluationsToSpecialize)
		&& (nbGoodAnticipations>0) && (nbBadAnticipations>0));    
   }
   
   /**
    * @param nbSuccessToProbe
    * @return boolean
    */
   public boolean doesAnticipateWell(byte nbSuccessToProbe) 
   {
	return ((nbGoodAnticipations>=nbSuccessToProbe) && (nbBadAnticipations==0));    
   }
   
   /**
    * @param nbFailuresToRemove
    * @return boolean
    */
   public boolean doesAnticipateBadly(byte nbFailuresToRemove) 
   {
	return ((nbBadAnticipations>=nbFailuresToRemove) && (nbGoodAnticipations==0));    
   }
   
   /**
    * @param initialReliability
    */
   public final void flushMarkers(double initialReliability) 
   {
	lastEvaluation	=	0; // ca se discute
	nbGoodAnticipations = 0;
	nbBadAnticipations = 0;
	initReliabilitiesArrays(initialReliability);
	//System.out.println("flushMarkers : " + this.toString(true)+ "***");    
   }
   
   /**
    * @param initialReliability
    */
   public final void freeMemories(double initialReliability) 
   {
    	nbGoodAnticipations = 0;
	nbBadAnticipations = 0;

	lastEvaluation = 0;

	lastPerceptionWhenAnticipatingWell = null;
	lastPerceptionWhenAnticipatingBadly = null;

	initReliabilitiesArrays(initialReliability);    
   }
   
   /**
    * @return int
    */
   public final int getLastEvaluation() 
   {
	return lastEvaluation;    
   }
   
   public final void resetLastEvaluation() 
   {
    	lastEvaluation = 0;    
   }
   
   public final void increaseLastEvaluation() 
   {
	lastEvaluation++;    
   }
   
   /**
    * @param nbEvaluationsToSpecialize
    * @param nbFailuresToRemove
    * @param nbSuccessToProbe
    * @return double
    */
   private double getCertainty(byte nbEvaluationsToSpecialize, byte nbFailuresToRemove, byte nbSuccessToProbe) 
   {
 	double max = nbGoodAnticipations + nbBadAnticipations;
 	if ((nbGoodAnticipations>0) && (nbBadAnticipations>0)) 	    	max = max / nbEvaluationsToSpecialize;
 	else if ((nbGoodAnticipations==0) && (nbBadAnticipations>0))	max = max / nbFailuresToRemove;
 	else if ((nbGoodAnticipations>0) && (nbBadAnticipations==0))    max = max / nbSuccessToProbe;

 	if (max>1.0) max = 1.0;  // la on a un fillrate;
	return max;    
   }
   
   /**
    * @param nbEvaluationsToSpecialize
    * @param nbFailuresToRemove
    * @param nbSuccessToProbe
    * @return double
    */
   public double getEvaluationLevel(byte nbEvaluationsToSpecialize, byte nbFailuresToRemove, byte nbSuccessToProbe) 
   {
	return ((1.0-getCertainty(nbEvaluationsToSpecialize, nbFailuresToRemove, nbSuccessToProbe))*getExplorationFactor());    
   }
   
   /**
    * @return boolean
    */
   public final boolean hasAlreadyBeenEvaluated() 
   {
	return ((nbGoodAnticipations+nbBadAnticipations)>=0);    
   }
   
   /**
    * @param lastPerception
    * @param learningRate
    */
   public final void addGoodAnticipation(Perception lastPerception, double learningRate) 
   {
	nbGoodAnticipations++;
	lastPerceptionWhenAnticipatingWell = new Perception(lastPerception);
	updateReliabilitiesIfWellAnticipating(lastPerception,learningRate);
	lastEvaluation = 0;    
   }
   
   /**
    * @param lastPerception
    * @param learningRate
    */
   public final void addBadAnticipation(Perception lastPerception, double learningRate) 
   {
	nbBadAnticipations++;
	lastPerceptionWhenAnticipatingBadly = new Perception(lastPerception);
	updateReliabilitiesIfBadlyAnticipating(lastPerception,learningRate);
	lastEvaluation = 0;    
   }
   
   /**
    * @return byte
    */
   public byte chooseIndexToMutSpec() 
   {
	byte choosenIndex = (byte)(-1);

	if (!condition.isSpecialized())
	    {
		byte indexArray[] = new byte[perceptionLength]; 
		ValuesDistribution distrib = buildValuesDistributionForUnreliabilityOfGeneralPerceptions(indexArray);
		distrib.updateToFit0_1 (0,1);
		choosenIndex = indexArray[distrib.computeIndexMax()];
	    }

	return choosenIndex;    
   }
   
   /**
    * @param lastGoodPerception
    * @param index
    */
   public void specialize(Perception lastGoodPerception, byte index) 
   {
	condition.setToken(index,lastGoodPerception.getToken(index));    
   }
   
   /**
    * The construction method has been changed here, it is a bit less efficient:
    * we copy the ReliabilitiesArrays and then we reinitialize them...
    * 
    * @param index
    * @param initialReliability
    * @param learningRate
    * @return AnticipatoryClassifier
    */
    public AnticipatoryClassifier computeGoodSpecialized(byte index, double initialReliability, double learningRate) 
    {
	if (condition.isDontCare(index))
	    {	
		AnticipatoryClassifier classifier = new AnticipatoryClassifier(this,false,initialReliability);
		classifier.addGoodAnticipation(lastPerceptionWhenAnticipatingWell,learningRate);
		classifier.specialize(lastPerceptionWhenAnticipatingWell, index);
		return classifier;
	    }
	return this;    
    }
   
   /**
    * Put a new dontCare token in the condition part in the most suitable place.
    * That place is indicated by the ReliabilitiesIfGeneral estimators.
    * 
    * @param initialReliability
    * @return AnticipatoryClassifier
    */
   public final AnticipatoryClassifier mutGenOne(double initialReliability) 
   {
	AnticipatoryClassifier newClassifier = new AnticipatoryClassifier(this,false,initialReliability);

	ValuesDistribution vd = buildValuesDistributionForReliabilitiesIfGeneral();
	short indexMax = vd.computeIndexMax();
	if (vd.getValue(indexMax)>0.5)
	    newClassifier.getCondition().generalize((byte)indexMax);

	if (newClassifier.getCondition().equals(getCondition()))    return this;
	else    return newClassifier;    
   }
   
   /**
    * @param indexArrayToBuild[]
    * @return Utils.ValuesDistribution
    */
   public final ValuesDistribution buildValuesDistributionForUnreliabilityOfGeneralPerceptions(byte indexArrayToBuild[]) 
   {
    	double tabReliability[] = new double[perceptionLength];
	byte nbGeneral = 0;

	for (byte i=0 ; i<perceptionLength ; i++)
	    {
		if (condition.getToken(i).isDontCare())
		    {
			tabReliability[i] = (1.0-reliabilitiesArray[i]);
			indexArrayToBuild[nbGeneral] = i;
			nbGeneral ++;
		    }
	    }	

	return (new ValuesDistribution(tabReliability, nbGeneral));    
   }
   /**
    * ------------- the following part is devoted to reliability exclusively... -------------------
    */
   
   /**
    * @return double[]
    */
   public final double[] getReliabilities() 
   {
	return reliabilitiesArray;    
   }
   
   /**
    * @return double[]
    */
   public final double[] getReliabilitiesIfGeneral() 
   {
	return reliabilitiesIfGeneralArray;    
   }
   
   /**
    * @param index
    * @param value
    */
   protected final void setReliabilitiesIfGeneral(byte index, double value) 
   {
	//System.out.println("old : " + reliabilitiesIfGeneralArray[index] + ", new : " + value);
	reliabilitiesIfGeneralArray[index] = value;    
   }
   
   /**
    * @param index
    * @return double
    */
   public final double getReliabilitiesIfGeneral(byte index) 
   {
	return reliabilitiesIfGeneralArray[index];    
   }
   
   /**
    * @return String
    */
   public final String getReliabilitiesIfGeneralString() 
   {
	String retour = new String("ReliabilitiesIfGeneral\t:(");
	for (byte j=0;j<getPerceptionLength();j++)
	    retour = retour + "[" + j + "] :" + getReliabilitiesIfGeneral(j) + " ";
	return retour + ")";    
   }
   
   /**
    * @return String
    */
   public final String getReliabilitiesString() 
   {
	String retour = new String("Reliabilities\t\t:(");
	for (byte j=0;j<getPerceptionLength();j++)
	    retour = retour + "[" + j + "] :" + reliabilitiesArray[j] + " ";
	return retour + ")";    
   }
   
   /**
    * @param initialReliability
    */
   private final void initReliabilitiesArrays(double initialReliability) 
   {
    	reliabilitiesArray = new double[perceptionLength];
	reliabilitiesIfGeneralArray = new double[perceptionLength];
	for (byte i=0; i<perceptionLength ; i++)
	    {
		reliabilitiesArray[i] = initialReliability;
		setReliabilitiesIfGeneral(i,initialReliability);
	    }    
   }
   
   /**
    * @return boolean
    */
   public final boolean isReadyToGeneralize() 
   {
	for (byte i=0 ; i<perceptionLength ; i++)
	    if (reliabilitiesIfGeneralArray[i]>0.5) return true;
	return false;    
   }
   
   /**
    * @param lastPerception@param learningRate
    */
   private final void updateReliabilitiesIfWellAnticipating(Perception lastPerception, double learningRate) 
   {
	if (lastPerceptionWhenAnticipatingBadly != null)

	    for (byte i=0 ; i<perceptionLength ; i++)
		{
		    reliabilitiesArray[i] = ((1.0-learningRate)*reliabilitiesArray[i]);    
		    if (lastPerception.getToken(i).equals(lastPerceptionWhenAnticipatingBadly.getToken(i)))
			reliabilitiesArray[i] += learningRate;
		}    
   }
   
   /**
    * @param lastPerception@param learningRate
    */
   private final void updateReliabilitiesIfBadlyAnticipating(Perception lastPerception, double learningRate) 
   {
	if (lastPerceptionWhenAnticipatingWell != null)
	    for (byte i=0 ; i<perceptionLength ; i++)
		{
		    reliabilitiesArray[i] = ((1.0-learningRate)*reliabilitiesArray[i]);    
		    if (lastPerception.getToken(i).equals(lastPerceptionWhenAnticipatingWell.getToken(i)))
			reliabilitiesArray[i] += learningRate;
		}    
   }
   
   /**
    * @param index@param learningRate
    */
   private void decreaseReliabilitiesIfGeneral(byte index, double learningRate) 
   {
	setReliabilitiesIfGeneral(index,((1.0-learningRate)*reliabilitiesIfGeneralArray[index]));    
   }
   
   /**
    * @param index@param learningRate
    */
   private void increaseReliabilitiesIfGeneral(byte index, double learningRate) 
   {
	setReliabilitiesIfGeneral(index,((1.0-learningRate)*reliabilitiesIfGeneralArray[index] + learningRate));    
   }
   
   /**
    * @param lastPerception
    * @param currentPerception
    * @param learningRate
    */
   public final void updateGeneralReliabilities(Perception lastPerception, Perception currentPerception, double learningRate) 
   {
	Perception specializedCondition = getCondition();
	Anticipation anticipation = getAnticipation();

	for (byte i=0 ; i<perceptionLength ; i++)
	    if (specializedCondition.isSpecialized(i))
		{
		    Perception generalCondition = new Perception(getCondition());
		    generalCondition.generalize(i);
		    if (generalCondition.matches(lastPerception) && !specializedCondition.matches(lastPerception)
			&& !anticipation.matches(currentPerception))
			{
			    decreaseReliabilitiesIfGeneral(i,learningRate);
			}

		    if (generalCondition.matches(lastPerception) && !specializedCondition.matches(lastPerception)
			&& anticipation.matches(currentPerception))
			{
			    increaseReliabilitiesIfGeneral(i,learningRate);
			}
		}    
   }
   
   /**
    * @return ValuesDistribution
    */
   public final ValuesDistribution buildValuesDistributionForReliabilities() 
   {
    	return (new ValuesDistribution(reliabilitiesArray, perceptionLength));    
   }
   
   /**
    * @return ValuesDistribution
    */
   public final ValuesDistribution buildValuesDistributionForReliabilitiesIfGeneral() 
   {
 	return (new ValuesDistribution(reliabilitiesIfGeneralArray, perceptionLength));    
   }
   
   /**
    * ------------- end of part devoted to reliability exclusively... -------------------
    */

    /**
    * 
    * @param verbose
    * @return String
    */
   public String toString(boolean verbose) 
   {
	String string = condition.toString() + " " + action.toString() + " " + anticipation.toString(); 

	if (verbose)
	    {
		string = string + "G:" + nbGoodAnticipations + " B:" + nbBadAnticipations + "\n";//;
		/*
		string = string + " " + getReliabilitiesString() + "\n";
 		string = string + " " + getReliabilitiesIfGeneralString() + "\n";
		*/
 		//string = string + " " + lastPerceptionWhenAnticipatingWell;
 		//string = string + " " + lastPerceptionWhenAnticipatingBadly;
		//string = string + " le:" + lastEvaluation + "(" + getExplorationFactor() + ")";
		//string = string + " cert:" + getCertainty();
		//string = string + " info:" + getEvaluationLevel();
		//string = string + " " + this.hashCode();
	    }
	return string;    
   }
}
