
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Classifier;

import Utils.StaticRandom;
/**
 * This class is used to implement standard classifiers such as ZCS classifiers
 * If you want to improve the speed of MACS, you can delete this class
 * and put all its methods in AnticipatoryClassifier.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
**/

public class BasicClassifier implements Comparable 
{
   protected byte perceptionLength;
   protected Perception condition;
   protected Action action;
   
   /**
    * @param classifier
    */
   public BasicClassifier(BasicClassifier classifier) 
   {
	action = new Action(classifier.action);
	condition = new Perception(classifier.condition);
	perceptionLength = condition.getLength();    
   }
   
   /**
    * @param perceptionForm
    * @param actionForm
    */
   public BasicClassifier(String perceptionForm, String actionForm) 
   {
	action = new Action(actionForm);
	condition = new Perception(perceptionForm);
	perceptionLength = Message.computeNbTokens(perceptionForm);    
   }
   
   public BasicClassifier() 
   {
    
   }
   
   /**
    * Returns a clone of the BasicClassifier
    * @return Object
    */
   public Object clone() 
   {
	return new BasicClassifier(this);    
   }
   
   /**
    * @param p
    * @return int
    */
   public int compareTo(Object p) 
   {
	int cond = condition.compareTo(((BasicClassifier)p).getCondition());
	if (cond !=0) return cond;
	else
	    {
		int ac = action.compareTo(((BasicClassifier)p).getAction());
		if (ac !=0) return ac;
	    }
	return 0;    
   }
   
   /**
    * @return byte
    */
   public final byte getPerceptionLength() 
   {
	return perceptionLength;    
   }
   
   /**
    * @return Action
    */
   public Action getAction() 
   {
	return action;    
   }
   
   /**
    * @return Perception
    */
   public Perception getCondition() 
   {
	return condition;    
   }
   
   /**
    * @return String
    */
   public String getActionString() 
   {
	return action.toString();    
   }
   
   /**
    * @return String
    */
   public String getPerceptionString() 
   {
	return condition.toString();    
   }
   
   /**
    * @param classifier
    * @param watchForAnticipation
    * @return boolean
    */
   public boolean isMoreGeneralThan(BasicClassifier classifier) 
   {
	return
	    (condition.isMoreGeneralThan(classifier.condition)
	     && ( action.equals(classifier.action) || action.isMoreGeneralThan(classifier.action) ));    
   }
   
   /**
    * @param classifier
    * @param watchForAnticipation
    * @return boolean
    */
   public boolean isLessGeneralThan(BasicClassifier classifier) 
   {
	return
	    (condition.isLessGeneralThan(classifier.condition)
	     && ( action.equals(classifier.action) || action.isMoreGeneralThan(classifier.action) ));    
   }
   
   /**
    * @param classifier
    * @return boolean
    */
   public boolean matches(BasicClassifier classifier) 
   {
	return
	    (condition.matches(classifier.condition)
	     && action.matches(classifier.action));    
   }
   
   /**
    * @param extCond
    * @return boolean
    */
   public boolean matchCondition(Perception extCond) 
   {
    	return (condition.matches(extCond));    
   }
   
   /**
    * @param act
    * @return boolean
    */
   public boolean matchAction(Action act) 
   {
    	return (action.matches(act));    
   }
   
   /**
    * @param extCond
    * @param act
    * @return boolean
    */
   public boolean matches(Perception extCond, Action act) 
   {
       return (matchCondition(extCond) && matchAction(act));    
   }
   
   /**
    * @param classifier
    * @return boolean true if the two objects are equal, else false
    */
   public boolean equals(Object classifier) 
   {
	return (condition.equals(((BasicClassifier)classifier).condition) 
		&& action.equals(((BasicClassifier)classifier).action));    
   }
   
   /**
    * @return Classifier
    * public  BasicClassifier mutGenOneBasic() 
    * {
    * BasicClassifier newClassifier = new BasicClassifier(this);
    * short indexMax = 
    * (short)(StaticRandom.nextDouble()*(double)(getCondition().getLength()));
    * newClassifier.getCondition().generalize((byte)indexMax);
    * return newClassifier;        
    * }
    * @param newExtCond
    */
   public void setCondition(Perception newExtCond) 
   {
	condition = newExtCond;    
   }
   
   /**
    * @param newAct
    */
   public void setAction(Action newAct) 
   {
	action = newAct;    
   }
   
   /**
    * @param verbose
    * @return String
    */
   public String toString(boolean verbose) 
   {
	return condition.toString() + " " + action.toString();    
   }
   
   /**
    * @return String
    */
   public String toString() 
   {
	return toString(false);    
   }
}
