
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Classifier;

import java.lang.reflect.Array;
import java.util.Vector;
import java.util.StringTokenizer;

/**
* This class is used to represent parts of classifiers.
* It is implemented by Action, Perception, Anticipation
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
**/

abstract class Message implements Comparable 
{
   protected byte nbSpecValues[];
   protected byte length;
   protected double specialization;
   protected boolean isSpecializationUpdated = false;
   protected Token tokenArray[];
   
   /**
    * @param source
    */
   protected Message(Message source) 
   {
	length = source.length;

	tokenArray = new Token[length];
	nbSpecValues = new byte[length];

	for (byte i=0; i<length; i++)
	    {
		tokenArray[i] = source.tokenArray[i];
		setNbSpecValues(i, source.getNbSpecValues(i));
	    }    
   }
   
   protected Message() 
   {
    
   }
   
   /**
    * @param string
    */
   protected final void setValue(String string) 
   {
	StringTokenizer tokenizer = new StringTokenizer (string, "|");
	String strTemp;
	for (byte i=0 ; i<length ; ++i)
	    {
		strTemp = tokenizer.nextToken();
		if (i==0)		strTemp = strTemp.substring(1);
		if (i==(length-1))	strTemp = strTemp.substring(0, strTemp.length()-1);
		tokenArray[i] = TokenRepository.getToken(strTemp);
	    }
	isSpecializationUpdated = false;    
   }
   
   /**
    * @return byte
    */
   public final byte getLength() 
   {
	return length;    
   }
   
   /**
    * @param i
    * @return Classifier.Token
    */
   public final Token getToken(byte i) 
   {
	return tokenArray[i];    
   }
   
   /**
    * @param i
    * @param tok
    */
   public final void setToken(byte i, Token tok) 
   {
	tokenArray[i] = tok;
	isSpecializationUpdated = false;    
   }
   
   /**
    * @param i
    * @param j
    */
   public final void setSpecializedToken(byte i, byte j) 
   {
	tokenArray[i] = TokenRepository.getToken(j);    
   }
   
   /**
    * @param message
    */
   protected final void setValue(Message message) 
   {
	for (byte i=0 ; i<length ; ++i)
	    tokenArray[i] = message.tokenArray[i];
	isSpecializationUpdated = false;    
   }
   
   /**
    * @param p
    * @return int
    */
   public int compareTo(Object p) 
   {
	for(byte i=0; i<length; i++)
	    {
		if (tokenArray[i].getValue()>((Message)p).tokenArray[i].getValue()) return -1;
		if (tokenArray[i].getValue()<((Message)p).tokenArray[i].getValue()) return 1;
	    }
	return 0;    
   }
   
   /**
    * @param message
    * @return boolean
    */
   public final boolean equals(Object message) 
   {
	for (byte i=0 ; i<length ; ++i)
	    if (!tokenArray[i].equals(((Message)message).tokenArray[i]))	return false;
	return true;    
   }
   
   /**
    * @param message
    * @param i
    * @return boolean
    */
   protected boolean equals(Message message, byte i) 
   {
	return (tokenArray[i].equals(message.tokenArray[i]));    
   }
   
   /**
    * @param message
    * @return boolean
    */
   public boolean matches(Message message) 
   {
	for (byte i=0 ; i<length ; ++i)
	    if (!tokenArray[i].matches(message.tokenArray[i]))	return false;
	return true;    
   }
   
   /**
    * @param message
    * @param i
    * @return boolean
    */
   protected final boolean matches(Message message, byte i) 
   {
	return (tokenArray[i].matches(message.tokenArray[i]));    
   }
   
   private final void updateSpecialization() 
   {
	byte nbSpec = 0;
	for (byte i=0 ; i<length ; ++i)
	    if (!tokenArray[i].isDontCare())
		nbSpec++;

	specialization = ((double)nbSpec/length);    
   }
   
   /**
    * @return double
    */
   public double getSpecialization() 
   {
	if (!isSpecializationUpdated)
	    {
		updateSpecialization();
		isSpecializationUpdated = true;
	    }
	return specialization;    
   }
   
   /**
    * @return byte
    */
   public final byte getIndexOfFirstDontCareToken() 
   {
	for (byte i=0 ; i<length ; ++i)
	    if (tokenArray[i].isDontCare())	return i;
	return (byte)(-1);    
   }
   
   /**
    * @return byte
    */
   public final byte getIndexOfFirstDontKnowToken() 
   {
	for (byte i=0 ; i<length ; ++i)
	    if (tokenArray[i].isDontKnow())	return i;

	return (byte)(-1);    
   }
   
   /**
    * @return boolean
    */
   protected boolean isGeneral() 
   {
	return (getSpecialization()==0);    
   }
   
   /**
    * @param i
    */
   public final void generalize(byte i) 
   {
	setToken(i,TokenRepository.dontCareToken);    
   }
   
   /**
    * @return boolean
    */
   public final boolean isSpecialized() 
   {
	return (getSpecialization()==1);    
   }
   
   /**
    * @param i
    * @return boolean
    */
   public final boolean isDontCare(byte i) 
   {
	return (tokenArray[i].isDontCare());    
   }
   
   /**
    * @param i
    * @return boolean
    */
   protected final boolean isDontChange(byte i) 
   {
	return (tokenArray[i].isDontChange());    
   }
   
   /**
    * @param i
    * @return boolean
    */
   public final boolean isDontKnow(byte i) 
   {
	return (tokenArray[i].isDontKnow());    
   }
   
   /**
    * @param message
    * @return boolean
    */
   public boolean isMoreGeneralThan(Message message) 
   {
	if (getSpecialization()>=message.getSpecialization())   return false;
	else
	    {
		boolean strictlyMoreFound = false;
		
		for (byte i=0 ; i<length ; ++i)
		    if (tokenArray[i].isMoreGeneralThan(message.tokenArray[i]))
			strictlyMoreFound = true;
		    else if (!tokenArray[i].equals(message.tokenArray[i]))
			return (false);

		return strictlyMoreFound;
	    }    
   }
   
   /**
    * @param message
    * @param i
    * @return boolean
    */
   protected final boolean isMoreGeneralThan(Message message, byte i) 
   {
	return (tokenArray[i].isMoreGeneralThan(message.tokenArray[i]));    
   }
   
   /**
    * @param message
    * @return boolean
    */
   public final boolean isLessGeneralThan(Message message) 
   {
	if (getSpecialization()<=message.getSpecialization())
	    return false;
	else
	    {
		boolean strictlyLessFound = false;
		
		for (byte i=0 ; i<length ; ++i)
		    if (tokenArray[i].isLessGeneralThan(message.tokenArray[i]))
			strictlyLessFound = true;
		    else if (!tokenArray[i].equals(message.tokenArray[i]))
			return (false);
		
		return strictlyLessFound;
	    }    
   }
   
   /**
    * @param message
    * @param i
    * @return boolean
    */
   protected final boolean isLessGeneralThan(Message message, byte i) 
   {
	return (tokenArray[i].isLessGeneralThan(message.tokenArray[i]));    
   }
   
   /**
    * @param message
    * @return boolean
    */
   public final boolean isEqualOrMoreGeneralThan(Message message) 
   {
	if (getSpecialization()>message.getSpecialization())
	    return false;
	else
	    {
		boolean strictlyMoreFound = false;
		
		for (byte i=0 ; i<length ; ++i)
		    if (tokenArray[i].isMoreGeneralThan(message.tokenArray[i]) || tokenArray[i].equals(message.tokenArray[i]))
			strictlyMoreFound = true;
		    else if (!tokenArray[i].equals(message.tokenArray[i]))
			return (false);

		return strictlyMoreFound;
	    }    
   }
   
   /**
    * @param message
    * @param i
    * @return boolean
    */
   protected final boolean isEqualOrMoreGeneralThan(Message message, byte i) 
   {
	return (tokenArray[i].isMoreGeneralThan(message.tokenArray[i]) || tokenArray[i].equals(message.tokenArray[i]));    
   }
   
   /**
    * @param message
    * @return boolean
    */
   protected final boolean isEqualOrLessGeneralThan(Message message) 
   {
	if (getSpecialization()<message.getSpecialization())
	    return false;
	else
	    {
		boolean strictlyLessFound = false;
		
		for (byte i=0 ; i<length ; ++i)
		    if (tokenArray[i].isLessGeneralThan(message.tokenArray[i]) || tokenArray[i].equals(message.tokenArray[i]))
			strictlyLessFound = true;
		    else if (!tokenArray[i].equals(message.tokenArray[i]))
			return (false);
		
		return strictlyLessFound;
	    }    
   }
   
   /**
    * @param message
    * @param i
    * @return boolean
    */
   protected final boolean isEqualOrLessGeneralThan(Message message, byte i) 
   {
	return (tokenArray[i].isLessGeneralThan(message.tokenArray[i]) || tokenArray[i].equals(message.tokenArray[i]));    
   }
   
   /**
    * @param msgForm
    * @param nbValuesArray
    */
   protected static final void readNbValues(String msgForm, byte[] nbValuesArray) 
   {
	StringTokenizer tokenizer = new StringTokenizer (msgForm, "|");
	String strTemp;

	byte i = 0;
	do
	    {
		strTemp = tokenizer.nextToken();
		if (i==0)
		    strTemp = strTemp.substring(1);
		if (!tokenizer.hasMoreElements())
		    strTemp = strTemp.substring(0, strTemp.length()-1);
		nbValuesArray[i] = Byte.valueOf(strTemp).byteValue();
		++i;
	    }
	while (tokenizer.hasMoreElements());    
   }
   
   /**
    * @param msgForm
    * @return byte
    */
   protected static final byte computeNbTokens(String msgForm) 
   {
	byte nbTokens = (byte)1;

	for (byte i=0; i < msgForm.length(); ++i)
	    if (msgForm.charAt(i) == '|')
		nbTokens++;

	return (nbTokens);    
   }
   
   /**
    * @param msgForm
    * @param nbValuesArray
    * @return byte
    */
   protected static final byte computeNbTokens(String msgForm, byte[] nbValuesArray) 
   {
	byte nbTokens = computeNbTokens(msgForm);	
	readNbValues (msgForm, nbValuesArray);

	return (nbTokens);    
   }
   
   /**
    * @return String
    */
   public final String getForm() 
   {
	String strTemp = "[";

	for (byte i=0; i<length; i++)
	    {
		strTemp = strTemp + getNbSpecValues(i);
		if (i != (length-1))
		    strTemp = strTemp + "|";
	    }

	strTemp = strTemp + "]";

	return strTemp;    
   }
   
   /**
    * @param index
    * @return byte
    */
   public final byte getNbSpecValues(int index) 
   {
	return nbSpecValues[index];    
   }
   
   /**
    * @param index
    * @param nbValues
    */
   protected final void setNbSpecValues(int index, byte nbValues) 
   {
	nbSpecValues[index] = nbValues;    
   }
   
   /**
    * @return String
    */
   public final String toString() 
   {
	String strTemp = "[";

	for (byte i=0; i<length; ++i)
	    {
		strTemp = strTemp + tokenArray[i].toString();
		if (i != (length-1))
		    strTemp = strTemp + "|";
	    }

	return strTemp + "]";    
   }
   
   /**
    * @return String
    */
   protected final String toStringShort() 
   {
	String strTemp = "";

	for (byte i=0; i<length; ++i)
	    {
		strTemp = strTemp + tokenArray[i].toString();
		if (i != (length-1))
		    strTemp = strTemp + "|";
	    }

	return strTemp;    
   }
}
