
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Environments;

import Utils.StaticRandom;
import Utils.Misc;
import java.util.ArrayList;
import java.util.Vector;
import java.util.TreeSet;

/**
 * This class is used to represent a standard Woods Environment 
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
 */

public class Woods extends Environment 
{
   protected static final boolean fixNbWalls = false;
   protected static final byte defaultXsize = 10;
   protected static final byte defaultYsize = 10;
   protected static final double defaultProbaWall = 0.30;
   protected static final short defaultNbWalls = (short ) 15;
   protected short xStart;
   protected short yStart;
   protected short nbSubGoals;
   protected short nbSubGoalsReached;
   protected short nbSubGoalsToReach;
   protected short nbRandomBits;
   protected short nbIrrelevantBits;
   protected short additionnalBits[];
   protected double reinforcementForFeeding = 1;
   protected byte action;
   protected boolean chooseStartAtRandom;
   protected short xSize;
   protected short ySize;
   protected WoodsLoader loader;
   protected WoodsGenerator generator;
   protected ArrayList actionList = new ArrayList();
   protected Cell initMap[][];
   protected Cell map[][];
   protected Agent agent;
   
   /**
    * @param fileName
    */
   public Woods(String fileName) 
   {
	agent = new Agent();
	if (!fileName.equals("none"))	loadWoods(fileName);
	else if (fixNbWalls)		generateWoods(defaultNbWalls);
	else				generateWoods(defaultProbaWall);
	//System.out.println(toString());    
   }
   
   public Woods() {}
   
   /**
    * Get the x size
    * @return short
    */
   public short getXSize() 
   {
	return xSize;    
   }
   
   /**
    * Get the y size
    * @return short
    */
   public short getYSize() 
   {
	return ySize;    
   }
   
   /**
    * @param fileName
    */
   protected void loadWoods(String fileName) 
   {
	setStart(-1,-1);

	loader = new WoodsLoader(this);
	Cell[][] tempomap = loader.fillWood(fileName);
	setMaps(tempomap);

	if (xStart == -1)	chooseStartAtRandom = true;
	else			chooseStartAtRandom = false;

	additionnalBits = new short[nbIrrelevantBits + nbRandomBits];

	fillActionList();

	updateRepresentativeRules();
	updateListOfPerceptions();

	//System.out.println(representativeRules);

	prepareForNewProblem();    
   }
   
   /**
    * @param probaWalls
    */
   protected void generateWoods(double probaWalls) 
   {
	generator = new WoodsGenerator(defaultXsize, defaultYsize);
	do
	    {
		generator.buildAdequateWoods(probaWalls);
		setMaps(generator.getMap());
	    }
	while (!isMarkov());
	
	finalizeWoodsGeneration();    
   }
   
   /**
    * @param nbWalls
    */
   protected void generateWoods(short nbWalls) 
   {
	generator = new WoodsGenerator(defaultXsize, defaultYsize);
	do
	    {
		generator.buildAdequateWoods(nbWalls);
		setMaps(generator.getMap());
	    }
	while (!isMarkov());
	
	finalizeWoodsGeneration();    
   }
   
   protected void finalizeWoodsGeneration() 
   {
	setStart(-1,-1);

	chooseStartAtRandom = true;
	nbSubGoalsToReach=(short)1;
	nbIrrelevantBits=(byte)0;
	nbRandomBits=(byte)0;
	additionnalBits = new short[nbIrrelevantBits + nbRandomBits];

	fillActionList();

	updateRepresentativeRules();
	updateListOfPerceptions();

	//System.out.println(representativeRules);

	prepareForNewProblem();    
   }
   
   protected void fillActionList() 
   {
	actionList.add("[0]");
	actionList.add("[1]");
	actionList.add("[2]");
	actionList.add("[3]");
	actionList.add("[4]");
	actionList.add("[5]");
	actionList.add("[6]");
	actionList.add("[7]");    
   }
   
   /**
    * @return java.util.Vector
    */
   public Vector prepareForNewProblem() 
   {
	input = null;
	firstStep = true;
	nbSubGoalsReached = 0;
	goalReached = false;
	initializeMap();
	chooseStart();
	initializeIrrelevantBits();
	
	action = -1;
	agent.goTo(xStart,yStart);

	Vector retour = new Vector();
	retour.add(agent);
	for (short j=0 ; j<ySize ; j++)
	    for (short i=0 ; i<xSize ; i++)
		retour.add(map[i][j]);
	return retour;    
   }
   
   protected void initializeIrrelevantBits() 
   {
	for (short i=0 ; i<nbIrrelevantBits ; i++)
	    if (StaticRandom.nextDouble()<0.5)
		additionnalBits[i] = (short)0;
	    else
		additionnalBits[i] = (short)1;    
   }
   
   protected void initializeRandomBits() 
   {
	for (short i=nbIrrelevantBits ; i<(nbRandomBits + nbIrrelevantBits) ; i++)
	    if (StaticRandom.nextDouble()<0.5)
		additionnalBits[i] = (short)0;
	    else
		additionnalBits[i] = (short)1;    
   }
   
   protected void chooseStart() 
   {
	//System.out.println(chooseStartAtRandom);
	if (chooseStartAtRandom)
	    {
		do
		    {
			xStart = (short)(StaticRandom.nextDouble()*xSize);
			yStart = (short)(StaticRandom.nextDouble()*ySize);
		    }
		while (!initMap[xStart][yStart].isFreeCell());
	    }    
   }
   
   protected void initializeMap() 
   {
	map = new Cell[xSize][ySize];

	for (short j=0 ; j<ySize ; j++)
	    for (short i=0 ; i<xSize ; i++)
		map[i][j] = initMap[i][j];    
   }
   
   public void changeCellValue(int i, int j, int value) 
   {
	map[i][j].setContent(value);
   }

   
   public String computeNewOutputs() 
   {
	return computeNewOutputs(true);    
   }
   
   /**
    * @param removeFoodWhenEaten
    */
   public String computeNewOutputs(boolean removeFoodWhenEaten) 
   {
	Coords save = new Coords(agent.getCoords());

	action = (byte)actionList.indexOf(input);
	agent.move(action, xSize, ySize);
	//System.out.println(agent.toString());

	reward = 0;
	if (map[agent.getXI()][agent.getYI()].isWallCell())
	    {
		agent.goTo(save);
	    }
	else if (map[agent.getXI()][agent.getYI()].isFoodCell())
	    {
		reward = reinforcementForFeeding;
		nbSubGoalsReached++;
		if (removeFoodWhenEaten) map[agent.getXI()][agent.getYI()].setContent(0);
	    }
	goalReached = (nbSubGoalsReached==nbSubGoals);
	initializeRandomBits();
	return computeOutput((short)agent.getXI(),(short)agent.getYI());    
   }
   
   protected String computeOutput(short x, short y)
   {
	String str = new String("[");
	str = str + map[x][y-1] + "|";
	str = str + map[x+1][y-1] + "|";
	str = str + map[x+1][y] + "|";
	str = str + map[x+1][y+1] + "|";
	str = str + map[x][y+1] + "|";
	str = str + map[x-1][y+1] + "|";
	str = str + map[x-1][y] + "|";
	str = str + map[x-1][y-1];

	for (short i=0; i<(nbIrrelevantBits+nbRandomBits) ;i++)
	    str = str + "|" + additionnalBits[i];

	return str + "]";
   }
   
   public String computeFirstOutputs() 
   {
	reward = 0;
	goalReached = (nbSubGoalsReached==nbSubGoals);
        return computeOutput((short)agent.getXI(),(short)agent.getYI());
   }
   
   /**
    * @return String
    */
   public String getOutputForm() 
   {
	String form = new String("[3|3|3|3|3|3|3|3");
	for (short i=0 ; i<nbIrrelevantBits ; i++)
		form = form + "|2";

	for (short i=0 ; i<nbRandomBits ; i++)
		form = form + "|2";

	return form +"]";    
   }
   
   /**
    * @return String
    */
   public String getInputForm() 
   {
	return "[8]";    
   }
   
   /**
    * @param x
    * @param y
    */
   protected void setStart(int x, int y) 
   {
	xStart = (short)x;
	yStart = (short)y;    
   }
   
   /**
    * @param nb
    */
   protected void setNbSubGoalsToReach(short nb) 
   {
	nbSubGoalsToReach = nb;    
   }
   
   /**
    * @param nb
    */
   protected void setNbIrrelevantBits(short nb) 
   {
	nbIrrelevantBits = nb;    
   }
   
   /**
    * @param nb
    */
   protected void setNbRandomBits(short nb) 
   {
	nbRandomBits = nb;    
   }
   
   /**
    * @param m[][] the map of cells
    */
   protected void setMaps(Cell m[][])
   {
	xSize = (short)m.length;
	ySize = (short)m[0].length;
	nbSubGoals = 0;

	initMap = new Cell[xSize][ySize];
	map = new Cell[xSize][ySize];

	for (short i=0 ; i<xSize ; i++)
	    for (short j=0 ; j<ySize ; j++)
		{
		    if (m[i][j].isFoodCell()) nbSubGoals++;
		    initMap[i][j] = m[i][j];
		    map[i][j] = m[i][j];
		}    
	for (short i=2 ; i<xSize-2 ; i++)
	    for (short j=2 ; j<ySize-2 ; j++)
		{
		    m[i][j].setPerception(computeOutput(i,j));
		    //		    System.out.println(i+":"+j+"=>"+map[i][j].getPerception());
		}
   }
   
   /**
    * @param x
    * @param y
    * @param act
    */
   protected void addRuleFor(short x, short y, String act) 
   {
	agent.goTo(x,y);
	String cond = computeOutput(x,y);

	setInput(act);
	output = computeNewOutputs(false);
	representativeRules = representativeRules + cond + " " + act + " " + output + "\n";
	//System.out.println(representativeRules);
	
	output = computeOutput(x,y);    
   }
   
   /**
    * @param nbBitsToFix
    * @param x
    * @param y
    */
   protected void fixAdditionnalBits(short nbBitsToFix, short x, short y) 
   {
	if (nbBitsToFix==(short)0)
	    {
		output = computeOutput(x,y);
		for (short i=0 ; i<actionList.size() ; i++)
		     addRuleFor(x,y,(String)actionList.get(i));
	    }
	else
	    {
		additionnalBits[nbBitsToFix-1] = 0;
		fixAdditionnalBits ((short)(nbBitsToFix-1),x,y);
		additionnalBits[nbBitsToFix-1] = 1;
		fixAdditionnalBits ((short)(nbBitsToFix-1),x,y);
	    }    
   }
   
   /**
    * @param x
    * @param y
    */
   protected void addRulesFor(short x, short y) 
   {
	fixAdditionnalBits((short)(nbIrrelevantBits+nbRandomBits),x,y);    
   }
   
   protected void updateRepresentativeRules() 
   {
	representativeRules = new String();
	
	for (short j=0 ; j<ySize ; j++)
	    for (short i=0 ; i<xSize ; i++)
		if (map[i][j].isFreeCell())    addRulesFor(i,j);    
   }
   
   /**
    * @param nbBitsToFix
    * @param x
    * @param y
    */
   protected void fixAdditionnalBits_perc(short nbBitsToFix, short x, short y) 
   {
	if (nbBitsToFix==(short)0)
	    {
		output = computeOutput(x,y);
		addPerceptionFor_2(x,y);
	    }
	else
	    {
		additionnalBits[nbBitsToFix-1] = 0;
		fixAdditionnalBits_perc ((short)(nbBitsToFix-1),x,y);
		additionnalBits[nbBitsToFix-1] = 1;
		fixAdditionnalBits_perc ((short)(nbBitsToFix-1),x,y);
	    }    
   }
   
   /**
    * @param x
    * @param y
    */
   protected void addPerceptionFor_2(short x, short y) 
   {
	output = computeOutput(x,y);
	listOfPerceptions = listOfPerceptions + output + "\n";
	coordsOfPerceptions = coordsOfPerceptions + "x:" + x + " y:" + y + "\n";    
   }
   
   /**
    * @param x
    * @param y
    */
   protected void addPerceptionFor_1(short x, short y) 
   {
	fixAdditionnalBits_perc((short)(nbIrrelevantBits+nbRandomBits), x, y);    
   }
   
   protected void updateListOfPerceptions() 
   {
	coordsOfPerceptions = new String();
	listOfPerceptions = new String();

	for (short j=0 ; j<ySize ; j++)
	    for (short i=0 ; i<xSize ; i++)
		if (!map[i][j].isWallCell())   addPerceptionFor_2(i,j);    
   }
   
   /**
    * @return boolean
    */
   protected boolean isMarkov() 
   {
	TreeSet situationSet = new TreeSet();

	for (short j=0 ; j<ySize ; j++)
	    for (short i=0 ; i<xSize ; i++)
		if (map[i][j].isFreeCell())
		    {
			computeOutput(i,j);
			if (situationSet.contains(output)) return false;
			else   situationSet.add(output);
		    }
	return true;    
   }
   
   /**
    * @return String
    */
   public String getStepInformation() 
   {
	String str = new String();

	char actionSymbol = 'S';
	if (action==0) actionSymbol = '0';
	else if (action==1) actionSymbol = '1';
	else if (action==2) actionSymbol = '2';
	else if (action==3) actionSymbol = '3';
	else if (action==4) actionSymbol = '4';
	else if (action==5) actionSymbol = '5';
	else if (action==6) actionSymbol = '6';
	else if (action==7) actionSymbol = '7';

	String strLine = new String("+");
	for (short i=0 ; i<xSize ; i++) strLine = strLine + " ";
	strLine = strLine + "+";

	str = strLine + "\n" + str;
	for (short j=0 ; j<ySize ; j++)
	    {
		str = str + " ";
		for (short i=0 ; i<xSize ; i++)
		    {
			if (agent.isIn(i,j))
			    str = str + actionSymbol;
			else if (map[i][j].isFreeCell())
			    str = str + " ";
			else if (map[i][j].isWallCell())
			    str = str + "#";
			else if (map[i][j].isFoodCell())
			    str = str + "X";	
		    }
		str = str + " ";
		if (j!=ySize) str = str + "\n";
	    }
	str = str + "\n" + strLine;

	str = Misc.putSideBySide("State : ", str, (short)12);
	return "Inputs :    Action     " + input + "\n" + "Outputs :   Reward     " + reward + "\n" + "            Perception " + output + "\n" + str;
   }
   
   /**
    * @return String
    */
   public String toString() 
   {
	String str = new String();

	for (short j=0 ; (j<ySize) ; j++)
	    {
		for (short i=0 ; (i<xSize) ; i++)
		    if (map[i][j].isFreeCell())
			str = str + " ";
		    else if (map[i][j].isWallCell())
			str = str + "#";
		    else if (map[i][j].isFoodCell())
			str = str + "X";
		str = str + "\n";
	    }
	return str;    
   }
}// end of Woods
