
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Environments;

import Utils.StaticRandom;
import Utils.Misc;
import java.util.Vector;

/**
 * This class is used to generate a random standard Woods Environment 
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
 */

public class WoodsGenerator 
{
   private byte sizeX;
   private byte sizeY;
   private byte xFood;
   private byte yFood;
   private Cell map[][];
   
   /**
    * @param sX
    * @param sY
    */
   public WoodsGenerator(byte sX, byte sY) 
   {
	sizeX = sX;
	sizeY = sY;

	map = new Cell[sizeX][sizeY];    
   }
   
   /**
    * @return Environments.Cell[][]
    */
   protected final Cell[][] getMap() 
   {
	return map;    
   }
   
   /**
    * @param p
    */
   private final void fillMap(double p) 
   {
	for (int i=0 ; i<sizeX ; i++)
	    for (int j=0 ; j<sizeY ; j++)
		{ 
		    if ((i==0) || (i==(sizeX-1)) || (j==0) || (j==(sizeY-1)) || (StaticRandom.nextDouble()<p))
			map[i][j] = new Cell(0,i,sizeY-j);
		    else
			map[i][j] = new Cell(3,i,sizeY-j);
		}

	do
	    {
		xFood = (byte)(StaticRandom.nextDouble()*(sizeX-2)+1);
		yFood = (byte)(StaticRandom.nextDouble()*(sizeY-2)+1);
	    }
	while (!map[xFood][yFood].isTrois());

	if (xFood>=(sizeX-1))	    xFood = (byte)(sizeX-2);
	if (yFood>=(sizeY-1))	    yFood = (byte)(sizeY-2);

	map[xFood][yFood].setContent(2);    
   }
   
   /**
    * @param nbWalls
    */
   private final void fillMap(short nbWalls) 
   {
	byte xWall, yWall;

	for (int i=0 ; i<sizeX ; i++)
	    for (int j=0 ; j<sizeY ; j++)
		{ 
		    if ( (i==0) || (i==(sizeX-1)) || (j==0) || (j==(sizeY-1)) )
			map[i][j] = new Cell(1,i,sizeY-j);
		    else
			map[i][j] = new Cell(3,i,sizeY-j);
		}

	for (short i=0 ; i<nbWalls ; i++)
	    {
		do
		    {
			xWall = (byte)(StaticRandom.nextDouble()*(sizeX-2)+1);
			yWall = (byte)(StaticRandom.nextDouble()*(sizeY-2)+1);
		    }
		while (!map[xWall][yWall].isTrois());

		if (xWall>=(sizeX-1))		    xWall = (byte)(sizeX-2);
		if (yWall>=(sizeY-1))		    yWall = (byte)(sizeY-2);

		map[xWall][yWall].setContent(1);
	    }

	do
	    {
		xFood = (byte)(StaticRandom.nextDouble()*(sizeX-2)+1);
		yFood = (byte)(StaticRandom.nextDouble()*(sizeY-2)+1);
	    }
	while (!map[xFood][yFood].isTrois());
	
	if (xFood>=(sizeX-1))	    xFood = (byte)(sizeX-2);
	if (yFood>=(sizeY-1))	    yFood = (byte)(sizeY-2);

	map[xFood][yFood].setContent(2);    
   }
   
   /**
    * @return boolean
    */
   private final boolean isAllright() 
   {
	boolean modif;
	short nbInaccessibleCells;

	do
	    {
		modif = false;
		nbInaccessibleCells = (short)0;
		for (byte j=0 ; j<sizeY ; j++)
		    for (byte i=0 ; i<sizeX ; i++)
			{
			    if ((map[i][j].isTrois()) && existsReachableAround(i,j) )
				{
				    modif = true;
				    map[i][j].setContent(0);
				}
			    if (map[i][j].isTrois()) nbInaccessibleCells++;
			}
	    }
	while (modif);

	return (nbInaccessibleCells==(short)0);    
   }
   
   /**
    * @param i
    * @param j
    * @return boolean
    */
   private final boolean existsReachableAround(byte i, byte j) 
   {
	return (
		map[i-1][j-1].isFreeCell()
		|| map[i-1][j].isFreeCell()
		|| map[i-1][j+1].isFreeCell()
		|| map[i][j-1].isFreeCell()
		|| map[i][j+1].isFreeCell()
		|| map[i+1][j-1].isFreeCell()
		|| map[i+1][j].isFreeCell()
		|| map[i+1][j+1].isFreeCell()
		|| map[i-1][j-1].isFoodCell()
		|| map[i-1][j].isFoodCell()
		|| map[i-1][j+1].isFoodCell()
		|| map[i][j-1].isFoodCell()
		|| map[i][j+1].isFoodCell()
		|| map[i+1][j-1].isFoodCell()
		|| map[i+1][j].isFoodCell()
		|| map[i+1][j+1].isFoodCell()
		);    
   }
   
   /**
    * @param p
    */
   protected final void buildAdequateWoods(double p) 
   {
	do
	    fillMap(p);
	while (!isAllright());    
   }
   
   /**
    * @param nbWalls
    */
   protected final void buildAdequateWoods(short nbWalls) 
   {
	do
	    fillMap(nbWalls);
	while (!isAllright());    
   }
   
   /**
    * @return String
    */
   public final String toString() 
   {
	String str = new String();

	for (byte j=0 ; j<sizeY ; j++)
	    {
		for (byte i=0 ; i<sizeY ; i++)
		    str = str + map[i][j];
		str = str + "\n";
	    }
	return str;    
   }
   
   /**
    * @param args
    */
   static void main(String[] args) 
   {
	for (int i=0 ; i<10 ; i++)
	    {
		WoodsGenerator gen = new WoodsGenerator((byte)10, (byte)10);
		gen.buildAdequateWoods(0.5);
		System.out.println(gen.toString());
	    }    
   }
}
