
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Experiments;

import Environments.Environment;
import Utils.Misc;
import MACS.ClassifierSystem;
import Utils.FilesWriter;
import java.util.StringTokenizer;

/**
* This class is used to record some performance measurements into files during experiments.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
**/

class ReportsManager 
{
    private int currentTimeStep;
    private int currentTrial;
    private int currentSimulation;
    private double cumulatedReward = 0;
    private String envInfos;
    private String systemInfos;
    private String initOutputDir;
    private int noEnvironment;
    private FilesWriter cumulatedRewardWriter;
    private FilesWriter nbClassifiersForEachTimeStep;
    private FilesWriter nbTimeStepsForEachTrial;
    private FilesWriter qantAverageForEachTimeStep;
    private FilesWriter specAverageForEachTimeStep;
    private FilesWriter knowledgePercentage;
    private FilesWriter trace;
    private Sequencer sequencer;
   
    /**
     * @param seq
     */
    ReportsManager(Sequencer seq) 
    {
	sequencer = seq;
	noEnvironment = (int)-1;
	initOutputDir = Main.parameters.outputDirectory + "/" + Main.parameters.experienceDesignation;
    }
   
    protected void setFree() 
    {
	sequencer = null;    
    }
   
    /**
     * @return String
     */
    private String getCurrentOutputDir() 
    {
	if (Main.parameters.differentDirectories)	return (initOutputDir + "_" + noEnvironment);
	else						return initOutputDir;    
    }
   
    private void storeClassifierSystem() 
    {
	try
	    {
		FilesWriter filesWriter = new FilesWriter(getCurrentOutputDir(),getCurrentOutputDir() + "/classifierSystem_" + currentSimulation +".res");
		filesWriter.outPrint(sequencer.getRuleBasedSystem().toString(true));
		filesWriter.flush();
		filesWriter.close();
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
   
    private void storeEnvironment() 
    {
	try
	    {
		FilesWriter filesWriter = new FilesWriter(getCurrentOutputDir(),getCurrentOutputDir() + "/environment_" + currentSimulation +".ini");
		filesWriter.outPrint(sequencer.getEnvironment().toString());
		filesWriter.flush();
		filesWriter.close();
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
   
    private void storeParameters() 
    {
	try
	    {
		FilesWriter filesWriter = new FilesWriter(getCurrentOutputDir(),getCurrentOutputDir() + "/config.ini");
		filesWriter.outPrint(Main.parameters.toString());
		filesWriter.flush();
		filesWriter.close();
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
   
    /**
     * @return String the file name
     */
    private String buildFileNameForPercentageOfKnowledge() 
    {
	return (getCurrentOutputDir() + "/simu" + currentSimulation + "_know.gp");    
    }
   
    /**
     * @return String the file name
     */
    private String buildFileNameForNbClassifiersForEachTimeStep() 
    {
	return (getCurrentOutputDir() + "/simu" + currentSimulation + "_nbcsf.gp");    
    }
   
    /**
     * @return String the file name
     */
    private String buildFileNameForSpecAverageForEachTimeStep() 
    {
	return (getCurrentOutputDir() + "/simu" + currentSimulation + "_specavg.gp");    
    }
   
    /**
     * @return String the file name
     */
    private String buildFileNameForQantAverageForEachTimeStep() 
    {
	return (getCurrentOutputDir() + "/simu" + currentSimulation + "_qantavg.gp");    
    }
   
    /**
     * @return String the file name
     */
    private String buildFileNameForNbTimeStepsForEachTrial() 
    {
	//(currentSimulation+Main.parameters.startSimuNumber)
	return (getCurrentOutputDir() + "/simu" + currentSimulation + "_nbtstrial.gp");    
    }
   
    /**
     * @return String the file name
     */
    private String buildFileNameForCumulatedReward() 
    {
	return (getCurrentOutputDir() + "/simu" + currentSimulation + "_cumulatedreward.gp");    
    }
   
    /**
     * @return String the file name
     */
    private String buildFileNameForTrace() 
    {	    
	return (getCurrentOutputDir() + "/simu" + currentSimulation + ".log");    
    }
   
    /**
     * @param reward the reward just received
     */
    public void storeCumulatedReward(double reward)
    {
	cumulatedReward += reward;
	try
	    {
		cumulatedRewardWriter.outPrint(cumulatedReward + "\n");
		cumulatedRewardWriter.flush();
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
    /**
     * @param nb
     */
    protected void prepareForTimeStep(int nb) 
    {
	currentTimeStep = nb;
	envInfos = sequencer.getEnvironment().getStepInformation();    
    }
   
    protected void endTimeStep() 
    {
	try
	    {
		if (Main.parameters.verboseTraceInFile || Main.parameters.verboseTraceOnTerm)
		    {
			systemInfos = sequencer.getRuleBasedSystem().getStepInformation();
 			if (Main.parameters.traceKnowledge)
 			    systemInfos = systemInfos + "Knowledge :         " + sequencer.getKnowledgePercentage() + " %\n";

			String verboseTrace = Misc.putSideBySide(envInfos, systemInfos, (short)50) + "\n";
			StringTokenizer stringTokenizer = new StringTokenizer(verboseTrace, "\n");
			int nb = stringTokenizer.countTokens();
			for (int i=0; i<(25-nb) ; i++) verboseTrace = verboseTrace + "\n";

			if (Main.parameters.verboseTraceInFile)
			    trace.outPrint(verboseTrace);
			if (Main.parameters.verboseTraceOnTerm)
			    System.out.print(verboseTrace);

			System.out.flush();
		    }
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
   
    /**
     * @param nb
     */
    protected void prepareForTrial(int nb) 
    {
	currentTrial = nb;    
    }
   
    /**
     * @param succeed
     */
    protected void endTrial(boolean succeed) 
    {
	try
	    {
		if (Main.parameters.reportsForGnuplot)
		    {
			nbClassifiersForEachTimeStep.outPrint((int)(sequencer.getRuleBasedSystem().getNbClassifiers()) + "\n");
 			if (Main.parameters.traceKnowledge) knowledgePercentage.outPrint(sequencer.getKnowledgePercentage() + "\n");
			String nbTS = new String();
			if ((currentTimeStep+1)<Main.parameters.nbStepsBeforeInterrupt)
			    //  nbTimeStepsForEachTrial.outPrint((int)(currentTimeStep+1) + "\n"); //question Pierre : pourquoi +1 ?
			    nbTimeStepsForEachTrial.outPrint(currentTimeStep + "\n");
			else
			    nbTimeStepsForEachTrial.outPrint("(failure)\n");			    
		    }
		if (Main.parameters.verboseTraceInFile)
		    {
			if (succeed) 
			    trace.outPrint("            ---- TRIAL " + currentTrial + " ACHIEVED IN " + currentTimeStep + " STEPS -----------------------------------------------\n\n");
			else
			    trace.outPrint("            ---- FAILED TO ACHIEVE TRIAL " + currentTrial + " -----------------------------------------------\n\n");
			/* why two times ???
			if (succeed)
			    trace.outPrint("            ---- TRIAL " + currentTrial + " ACHIEVED IN " + currentTimeStep + " STEPS -----------------------------------------------\n\n");
			else
			    trace.outPrint("            ---- FAILED TO ACHIEVE TRIAL " + currentTrial + " -----------------------------------------------\n\n");
			*/
		}
		if (Main.parameters.tinyTraceOnTerm)
		    {
			if (succeed)	System.out.print(currentTimeStep + " ");
			else		System.out.print("(failure)");
		    }
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
   
    /**
     * @param nb
     */
    protected void prepareForSimulation(int nb) 
    {
	currentSimulation = nb;
	if (sequencer.environmentJustChanged())	    noEnvironment=noEnvironment+1;
	try
	    {
		if (Main.parameters.traceCumulatedReward)
		    {
			cumulatedReward = 0;
			cumulatedRewardWriter = new FilesWriter(getCurrentOutputDir(),buildFileNameForCumulatedReward());
		    }
		if ( (Main.parameters.reportsForGnuplot || Main.parameters.verboseTraceInFile)
		     && ( (sequencer.environmentJustChanged() && Main.parameters.differentDirectories)
			  || ((currentSimulation==0) && (!Main.parameters.differentDirectories)) ) )
		    {
			if (Main.parameters.storeEnvironment) storeEnvironment();
		    }
		if (Main.parameters.reportsForGnuplot)
		    {
			nbClassifiersForEachTimeStep = new FilesWriter(getCurrentOutputDir(),buildFileNameForNbClassifiersForEachTimeStep());
			nbTimeStepsForEachTrial = new FilesWriter(getCurrentOutputDir(),buildFileNameForNbTimeStepsForEachTrial());
			if (Main.parameters.traceKnowledge) knowledgePercentage = new FilesWriter(getCurrentOutputDir(),buildFileNameForPercentageOfKnowledge());
		    }
		if (Main.parameters.verboseTraceInFile)
		    trace = new FilesWriter(getCurrentOutputDir(),buildFileNameForTrace());
		    
		if (Main.parameters.tinyTraceOnTerm)
		    {
			if (sequencer.environmentJustChanged())
			    System.out.print ("\n\n" + Utils.Misc.putSideBySide("Environment : ", sequencer.getEnvironment().toString(), (short)16));
			    System.out.print ("Simulation " + currentSimulation + " :\t");
		    }
		if (Main.parameters.verboseTraceInFile)
		    trace.outPrint("==== SIMULATION " + currentSimulation + " ==== \n\n");
		if (Main.parameters.verboseTraceOnTerm)
		    System.out.print("==== SIMULATION " + currentSimulation + " ==== \n\n");
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
   
    protected void endSimulation() 
    {
	try
	    {
		if (Main.parameters.traceCumulatedReward)
		    {
			cumulatedRewardWriter.close();
		    }
		if (Main.parameters.storeFinalSet)
		    storeClassifierSystem();
		if (Main.parameters.tinyTraceOnTerm)
		    System.out.print ("\n");
		if (Main.parameters.reportsForGnuplot)
		    {
			nbClassifiersForEachTimeStep.close();
			nbTimeStepsForEachTrial.close();
			if (Main.parameters.traceKnowledge) knowledgePercentage.close();
		    }
		if (Main.parameters.verboseTraceInFile)
		    {
			storeClassifierSystem();
			trace.close();
		    }
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
    }
}
