
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Experiments;

import Environments.Environment;
import MACS.RuleBasedSystem;
import MACS.ClassifierSystem;
import Utils.StaticRandom;
import Utils.Chrono;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.BufferedReader;
import java.lang.reflect.Constructor;

/**
 * The sequencer lets alternatively the environment and the agent compute their 
 * new outputs. It works as the following :
 * 1  - The environment computes its first outputs and given reward
 * 2a - The agent (RuleBasedSystem) gets the outputs and reward and learns what it 
 * can
 * 2b - The agent acts (i.e? sets its own output)
 * 2c - The environments computes its new output and reward according to the 
 * action computed by the agent
 * 2d - Go to 2a until the agent is allowed to act
 * 3  - Perform last step
 * NB : 2* is a step as we defined it. The agent perceives at once perceptions and 
 * reward, and then learns and acts.

* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */
class Sequencer 
{
   private final short NB_SIMUS_PER_ENV = (short)10;
   private boolean environmentChanged = true;
   private int nbFailures;
   private double knowledge;
   private Environment environment;
   private static RuleBasedSystem system;
   private ClassifierSystem controlRules;
   private FileWriter reportFile = null;
   private ReportsManager reportsManager;
   
   /**
    * private double initEpsilonGreedy;
    * @param env
    * @param syst
    */
   Sequencer(Environment env, RuleBasedSystem syst) 
   {
	environment = env;
	system = syst;
	reportsManager = new ReportsManager(this);
	controlRules = new MACS.ClassifierSystem(environment.getRulesForm(), environment.getRepresentativeRules());
   }
   
   /**
    * @return Experiments.RuleBasedSystem
    */
   protected RuleBasedSystem getRuleBasedSystem() 
   {
	return system;    
   }
   
   /**
    * @return Environments.Environment
    */
   public Environment getEnvironment() 
   {
	return environment;    
   }
   
   /**
    * @param step
    */
   protected void takeStep(int step, int globalstep) 
   {
	environment.takeStep();
	if (globalstep==Main.parameters.nbStepsBeforeChange)
	    {
		if (Main.parameters.environmentConfig.equals("/mnt/home/sigaud/osd/MACS/CurrentMACS/Config/EnvironmentsConfig/blockingBis.cfg"))
		{
		    System.out.println("OK1");
		    environment.changeCellValue(10,5,1);
		    environment.changeCellValue(2,5,0);
		}
		if (Main.parameters.environmentConfig.equals("/mnt/home/sigaud/osd/MACS/CurrentMACS/Config/EnvironmentsConfig/shortcutBis.cfg"))
		    {
			System.out.println("OK2");
			environment.changeCellValue(10,5,0);
		    }
		((MACS.ModularClassifierSystem)system).getTraceManager().setOrderedPerceptions(environment.getListOfPerceptions());
		((MACS.ModularClassifierSystem)system).getTraceManager().setCoordsOfOrderedPerceptions(environment.getCoordsOfPerceptions());
	    }
	reportsManager.prepareForTimeStep(step);

	//if (globalstep==Main.parameters.nbStepsBeforeChange+500) MACS.ClassifierSystemParameters.maximizeRewardFirst = true;

	//System.out.println("avant run");
	if (Main.parameters.traceCumulatedReward) reportsManager.storeCumulatedReward(environment.getReward());
 	environment.setInput(system.run(environment.getReward(),
					environment.getOutput(),
					environment.isGoalReached()));
	//System.out.println("aprs run");
	checkForKnowledgePercentage(step);
	reportsManager.endTimeStep();
   }
   
   /**
    * @param nbSteps
    * @return int
    */
   protected int runTrial(int nbSteps) 
   {
	environment.prepareForNewProblem();
	system.prepareForNewProblem();

	int i=0;
	while ((i<nbSteps) && (!environment.isGoalReached()))
	    {
		takeStep(i,i);
		i++;
	    }
	
	if (environment.isGoalReached())   return i;
	else
	    {
		nbFailures ++;
		return(-1);
	    }    
   }
   
   /**
    * @param nbSteps
    */
   private void runSimulation(int nbSteps) 
   {
       environment = Main.createEnvironment();
	if (Main.parameters.system.equals("MACS"))
	    {
		system = new MACS.ModularClassifierSystem(environment.getRulesForm(), Main.loadInitString(), Main.parameters.systemConfig);
		controlRules = new MACS.ClassifierSystem(environment.getRulesForm(), environment.getRepresentativeRules());
		((MACS.ModularClassifierSystem)system).getTraceManager().setOrderedPerceptions(environment.getListOfPerceptions());
		((MACS.ModularClassifierSystem)system).getTraceManager().setCoordsOfOrderedPerceptions(environment.getCoordsOfPerceptions());
	    }
	       
	environment.prepareForNewProblem();
	system.prepareForNewProblem();
	try
	    {
		int nbStepsToComplete = 0;
		int trialNumber = 0;
		reportsManager.prepareForTrial(trialNumber);

		for (int i=0 ; i<nbSteps ; i++)
		    {
			takeStep(nbStepsToComplete,i);
			nbStepsToComplete++;
			if (environment.isGoalReached())
			    {
				reportsManager.endTrial(nbStepsToComplete!=-1);
				trialNumber++;
				reportsManager.prepareForTrial(trialNumber);
				nbStepsToComplete = 0;
				environment.prepareForNewProblem();
				system.prepareForNewProblem();
			    }
		    }
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
   }
   
   /**
    * @param nbTrials
    * @param nbSteps
    */
   private void runSimulation(int nbTrials, int nbSteps) 
   {
       environment = Main.createEnvironment();
	if (Main.parameters.system.equals("MACS"))
	    {
		system = new MACS.ModularClassifierSystem(environment.getRulesForm(), Main.loadInitString(), Main.parameters.systemConfig);
		((MACS.ModularClassifierSystem)system).getTraceManager().setOrderedPerceptions(environment.getListOfPerceptions());
		((MACS.ModularClassifierSystem)system).getTraceManager().setCoordsOfOrderedPerceptions(environment.getCoordsOfPerceptions());
	    }

	controlRules = new MACS.ClassifierSystem(environment.getRulesForm(), environment.getRepresentativeRules());
	try
	    {
		int nbStepsToComplete = 0;
		for (int i=0; ((i<nbTrials) && (nbStepsToComplete!=-1)) ; i++)
		    {
			reportsManager.prepareForTrial(i);
			nbStepsToComplete = runTrial(nbSteps);
			reportsManager.endTrial(nbStepsToComplete!=-1);
		    }
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}    
   }
   
   /**
    * @param noSimu
    */
   private void changeEnvironment(short noSimu) 
   {
	if ( (NB_SIMUS_PER_ENV!=0) && (((noSimu+1)%NB_SIMUS_PER_ENV)==0) )
	    {
		environmentChanged = true;
		environment = Main.createEnvironment();
	    }
	else
	    environmentChanged = false;    
   }
   
   /**
    * @return boolean
    */
   protected boolean environmentJustChanged() 
   {
	return environmentChanged;    
   }
   
   /**
    * @param nbSimulations
    * @param nbTrials
    * @param nbSteps
    */

   public void run(int nbSimulations, int nbTrials, int nbSteps) 
   {
	//System.out.println("\n");
	try
	    {
		for (int i=0 ; i<nbSimulations ; i++)
		    {
			reportsManager.prepareForSimulation(i+Main.parameters.startSimuNumber);
			runSimulation(nbTrials, nbSteps);
			reportsManager.endSimulation();
			changeEnvironment((short)i);
		    }
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}

	//System.out.println("\n" + Main.parameters.nbSimulations + " simulations completed : " + nbFailures + " failures\n");    
   }
   
   /**
    * @param nbSimulations
    * @param nbSteps
    */
   public void run(int nbSimulations, int nbSteps) 
   {
	//System.out.println("\n");
	try
	    {
		for (int i=0 ; i<nbSimulations ; i++)
		    {
			reportsManager.prepareForSimulation(i+Main.parameters.startSimuNumber);
			runSimulation(nbSteps);
			reportsManager.endSimulation();		
			changeEnvironment((short)i);
		    }
	    }  catch (Exception e) {e.printStackTrace();System.exit(-1);}
	
	//System.out.println("\n" + Main.parameters.nbSimulations + " simulations completed");    
   }
   
   /**
    * @return double
    */
   protected double getKnowledgePercentage() 
   {
	return knowledge;    
   }
   
   /**
    * @param step
    */
   protected void checkForKnowledgePercentage(int step) 
   {
 	if ((Main.parameters.knowledgeComputationFrequency!=0) && (Main.parameters.traceKnowledge)
	    && ((step%Main.parameters.knowledgeComputationFrequency)==0))
	    knowledge = system.checkForConformityWith(controlRules);    
   }
}
