
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.*;
import Classifier.Perception;
import Classifier.Action;


/**
 * The AnticipatedSituationMap class is used to store temporarily
 * the list of anticipated situations computed from each perception
 * through the model of transitions.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */
final class Anticipator extends TreeMap
{
    public static boolean DEBUG = false;//true;//

    protected ClassifierSystem modelOfTransitions;   
    /**
     * @param modTrans
     */
    protected Anticipator(ClassifierSystem modTrans) 
    {
	modelOfTransitions = modTrans;    
    }
    /**
     * This methods computes for each possible action the list of anticipated 
     * situations 
     * and the associated information gains, and it fills tables with it:
     * - retour[0] = actions (contains the list of possible actions)
     * - retour[1] = anticipationSet (contains the list of anticipations from 
     * classifiers)
     * - retour[2] = infoGains (contains the list of infogains corresponding to each 
     * classifier)
     * If the value has already been computed, it does not compute it again.
     * 
     * @param currentPerception
     * @return java.util.Vector[]
     */
    public Vector[] getPossibilityTables(Perception currentPerception) 
    {
	Vector[] retour = (Vector[]) get(currentPerception);
	if (retour==null) 
	    {
		//if (DEBUG) System.out.println(" in getPossibilityTables");
		return getTables(currentPerception);
	    }
	else return retour;
    }
    public void buildPossibilityTables(Perception currentPerception) 
    {
	//if (DEBUG) System.out.println("in buildPossibilityTables : " +currentPerception);
	put(currentPerception, getTables(currentPerception));
    }
    private Vector[] getTables(Perception currentPerception) 
    {
	Vector[] retour = new Vector[3];
	retour[0] = new Vector();// actions
	retour[1] = new Vector();// anticipationSet
	retour[2] = new Vector();// infoGains

	for (byte i=0 ; i<modelOfTransitions.getNbActions(); i++)
	    {
		Action action = modelOfTransitions.getAction(i);
		Vector anticip = modelOfTransitions.computeAnticipatedSituations(currentPerception, action);

		if (anticip.size()==0)
		    {
			retour[0].add(action);
			retour[1].add(null);
			retour[2].add(new Double(1.0));
		    }
		else
		    {
			for (int j=0 ; j<anticip.size() ; j++)
			    {
				retour[0].add(action);
				retour[1].add(((CouplePF)anticip.get(j)).getFirst());
				retour[2].add(new Double(((CouplePF)anticip.get(j)).getSecond()));
			    }
		    }
	    }
	return retour;
    }
}
