
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.Vector;
import Utils.StaticRandom;
import Utils.ValuesDistribution;
import Classifier.Action;
import Classifier.Perception;
import Classifier.BasicClassifier;
import Classifier.AnticipatoryClassifier;

/**
 * This class is intended to store a list of AnticipatoryClassifier.
 * It is extended by ClassifierSystem.
 * It defines a lot of useful low level functionalities.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 **/
public class AnticipatoryClassifierList extends ClassifierList 
{
   /**
    * copy constructor
    * @param cset
    */
   public AnticipatoryClassifierList(AnticipatoryClassifierList cset) 
   {
	super((ClassifierList) cset);    
   }
   
   public AnticipatoryClassifierList() {}
   
   /**
    * @param cl
    */
   protected final void addClassifier(AnticipatoryClassifier cl) 
   {
	addClassifier(cl, true);    
   }
   
   public final void freeMemories() 
   {
	for (short i=0; i<getNbClassifiers() ; i++)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		acl.freeMemories(ClassifierSystemParameters.initialReliability);
	    }    
   }
   
   /**
    * @return int
    */
   public int getNbWellAnticipatingClassifiers() 
   {
	int nb=0;
	for (short i=0 ; i<getNbClassifiers() ; i++)
	    if (((AnticipatoryClassifier)getClassifier(i)).doesAnticipateWell(ClassifierSystemParameters.nbSuccessToProbe))
		nb++;

	return nb;    
   }
   
   /**
    * @return boolean
    */
   protected final boolean hasBeenEvaluatedEnoughForMutspec() 
   {
	for (short i=0; i<getNbClassifiers(); ++i)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		if (!(acl.doesOscillate(ClassifierSystemParameters.nbEvaluationsToSpecialize) || acl.doesAnticipateBadly(ClassifierSystemParameters.nbFailuresToRemove) || acl.doesAnticipateWell(ClassifierSystemParameters.nbSuccessToProbe)))
		    return false;
	    }
	return true;    
   }
   
   /**
    * @return boolean
    */
   protected final boolean hasBeenEvaluatedEnoughForGeneralization() 
   {
	boolean retour = false;
	for (short i=0; i<getNbClassifiers(); ++i)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		if (!acl.doesAnticipateWell(ClassifierSystemParameters.nbSuccessToProbe))	return false;
		else if (acl.isReadyToGeneralize())	retour=true;// nb : il se peut qu'on sorte par false avant...
	    }
	return retour;    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return boolean
    */
   protected final boolean contains(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	for (short i=0; i<size(); i++)
	    if (((AnticipatoryClassifier)getClassifier(i)).equals(newClassifier, watchForAnticipation))
		return true;
	return false;    
   }
   
   /**
    * @param newClassifier
    * @return boolean
    */
   public final boolean contains(AnticipatoryClassifier newClassifier) 
   {
	return contains(newClassifier, true);    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return boolean
    */
   protected final boolean containsClassifierMatching(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    if (((AnticipatoryClassifier)getClassifier(i)).matches(newClassifier, watchForAnticipation))
		return true;
	return false;    
   }
   
   /**
    * @param situation
    * @return boolean
    */
   protected final boolean containsWellAnticipatingClassifierMatching(Perception situation) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    if (((AnticipatoryClassifier)getClassifier(i)).doesAnticipateWell(ClassifierSystemParameters.nbSuccessToProbe) 
		&& getClassifier(i).matchCondition(situation))
		return true;

	return false;    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return boolean
    */
   protected final boolean containsLessGeneralThan(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    if (((AnticipatoryClassifier)getClassifier(i)).isLessGeneralThan(newClassifier, watchForAnticipation))
		return true;
	
	return false;    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return boolean
    */
   protected final boolean containsMoreGeneralThan(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    if (((AnticipatoryClassifier)getClassifier(i)).isMoreGeneralThan(newClassifier, watchForAnticipation))
		return true;
	
	return false;    
   }
   
   /**
    * @param classifier
    * @return Classifier.Perception
    */
   protected final Perception checkForIncompatibilityWith(AnticipatoryClassifier classifier) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		if ( (acl!=classifier)
		     && (!acl.getAnticipation().isCompatibleWith(classifier.getAnticipation())) )
		    return (new Perception((Perception)classifier.getCondition()));
	    }
	return null;    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return short
    */
   protected final short getNbMoreSpecializedThan(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	short nb = 0;

	for (short i=0; i<getNbClassifiers(); i++)
	    if (((AnticipatoryClassifier)getClassifier(i)).isLessGeneralThan(newClassifier, watchForAnticipation))
		nb++;
	
	return nb;    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return boolean
    */
   protected final boolean containsEqualOrLessGeneralThan(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		if ( acl.isLessGeneralThan(newClassifier, watchForAnticipation)
		     || acl.equals(newClassifier, watchForAnticipation) )
		    return true;
	    }
	
	return false;    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return boolean
    */
   protected final boolean containsEqualOrMoreGeneralThan(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		if ( acl.isMoreGeneralThan(newClassifier, watchForAnticipation)
		     || acl.equals(newClassifier, watchForAnticipation) )
		    return true;
	    }
	
	return false;    
   }
   
   /**
    * @param newClassifier
    * @param watchForAnticipation
    * @return boolean
    */
   protected final boolean containsComparableWith(AnticipatoryClassifier newClassifier, boolean watchForAnticipation) 
   {
	for (short i=0; i<getNbClassifiers(); i++)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		if ( acl.isMoreGeneralThan(newClassifier, watchForAnticipation)
		     || acl.isLessGeneralThan(newClassifier, watchForAnticipation)
		     || acl.equals(newClassifier, watchForAnticipation) )
		    return true;
	    }
	
	return false;    
   }
   
   /**
    * @param perception
    * @param act
    * @return AnticipatoryClassifierList
    */
   protected AnticipatoryClassifierList selectClassifiersMatching(Perception perception, Action act) 
   {
	AnticipatoryClassifierList matchingSet  = new AnticipatoryClassifierList();

	for (short i=0 ; i<getNbClassifiers() ; i++)
	    if (getClassifier(i).matches(perception,act))
		matchingSet.addClassifier(getClassifier(i), false);
	
	return matchingSet;    
   }
   
   /**
    * --------------------------------------------------------------------------------
    * @param concernedSets
    * @return MACS.AnticipationList
    */
   private static AnticipationList buildAnticipationList(java.util.Vector concernedSets) 
   {
	AnticipationList anticipationSet = new AnticipationList();
	for (short i=0; i<concernedSets.size() ; i++)
	    {
		AnticipatoryClassifierList acs = (AnticipatoryClassifierList)concernedSets.get(i);
		anticipationSet.add(((AnticipatoryClassifier)acs.getClassifier((short)0)).getAnticipation());
	    }
	return anticipationSet;
   }
   
   /**
    * Creates sets of classifiers with the correct action,
    * that do not match the anticipation set,
    * sorted by perception
    * 
    * @param concernedSets
    * @param lastAction
    * @return java.util.Vector
    */
   protected final Vector buildIncompatibleSets(java.util.Vector concernedSets, Action lastAction) 
   {
	AnticipationList anticipationList = buildAnticipationList(concernedSets);
	Vector incompatibleSets = new Vector();
	for (short i=0; i<anticipationList.size() ; i++) 
	    incompatibleSets.addElement(new AnticipatoryClassifierList());
	    
	for (short i=0; i<getNbClassifiers() ; i++)
	    {
		AnticipatoryClassifier classifier = (AnticipatoryClassifier)getClassifier(i);
		if (classifier.getAction().equals(lastAction))
		    {
			short index = (short)(anticipationList.getIndexOfFirstIncompatibleAnticipation(classifier.getAnticipation()));
			if (index!=-1)
			    {
				((AnticipatoryClassifierList)incompatibleSets.get(index)).addClassifier(classifier, false);
			    }
		    }
	    }
	return incompatibleSets;    
   }
   
   /**
    * --------------------------------------------------------------------------------
    * -------
    * @param lastAction
    * @param currentPerception
    * @return java.util.Vector
    */
   protected final Vector buildAnticipatoryConcernedSets(Action lastAction, Perception currentPerception) 
   {
	Vector concernedSets = new Vector();
	AnticipationList anticipationList = new AnticipationList();
	    
	for (short i=0; i<getNbClassifiers(); i++)
	    {
		AnticipatoryClassifier classifier = (AnticipatoryClassifier)getClassifier(i);
		if (classifier.getAction().equals(lastAction) && classifier.matchAnticipation(currentPerception))
		    {
			if (!anticipationList.contains(classifier.getAnticipation()))
			    {
				anticipationList.add(classifier.getAnticipation());
				concernedSets.addElement(new AnticipatoryClassifierList());
				((AnticipatoryClassifierList)concernedSets.get(concernedSets.size()-1)).addClassifier(classifier);
			    }
			else
			    {
				short index = (short)anticipationList.getIndexOfAnticipation(classifier.getAnticipation());
				((AnticipatoryClassifierList)concernedSets.get(index)).addClassifier(classifier);
			    }
		    }
	    }
	return (concernedSets);    
   }
}
