
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.awt.*;
import java.util.*;
import java.lang.*;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.FileWriter;

import Utils.StaticRandom;

import Classifier.BasicClassifier;
import Classifier.AnticipatoryClassifier;
import Classifier.Transition;
import Classifier.Perception;
import Classifier.Action;
/**
 * This class is used to represent a list of anticipatory classifiers
 * and contains methods used in LCSs.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */

public final class ClassifierSystem extends AnticipatoryClassifierList 
{
    public static boolean DEBUG = false;//true;//
   
    public static boolean localDEBUG = false;//true;//
   
    protected static final boolean TIME = false;//true;//
    protected static final boolean TRACE = false;//true;//
    protected String perceptionForm;
    protected String actionForm;
    private short nbActions = 0;
    private long time = 0;
   
    public BasicClassifier lastActiveClassifier;
    private Action possibleActionsArray[];
    private AnticipationLearningManager anticipationLearningManager;
    private SpecializationManager specializationManager;
    private ConditionCoveringManager conditionCoveringManager;
    private GeneralizationManager generalizationManager;
    private ClassifierSystemLoader loader;
    private ClassifierSystemTraceManager traceManager;
    private PerceptionList knownPerceptions;
    private MACSActionSelector perceptionListPointer;// this one is not necessary
   
    /**
     * @param csyst
     */
    public ClassifierSystem(ClassifierSystem csyst) 
    {
	super((AnticipatoryClassifierList) csyst);

	perceptionForm = new String(csyst.perceptionForm);
	actionForm = new String(csyst.actionForm);

	loader = new ClassifierSystemLoader(this);
	fillPossibleActionsArray();

	time = csyst.time;

	anticipationLearningManager = new AnticipationLearningManager(this);
	specializationManager = new SpecializationManager(this);
	conditionCoveringManager = new ConditionCoveringManager(this);
	traceManager = new ClassifierSystemTraceManager(csyst.traceManager, this);
	generalizationManager = new GeneralizationManager(this);    
    }
   
    /**
     * @param rulesForm
     * @param initString
     */
    public ClassifierSystem(String rulesForm, String initString) 
    {
	loader = new ClassifierSystemLoader(this);
	loader.setForms(rulesForm);
	fillPossibleActionsArray();

	anticipationLearningManager = new AnticipationLearningManager(this);
	specializationManager = new SpecializationManager(this);
	conditionCoveringManager = new ConditionCoveringManager(this);
	traceManager = new ClassifierSystemTraceManager(this);
	generalizationManager = new GeneralizationManager(this);

	loader.fillFromString(initString);    
    }
   
    /**
     * @param p
     */
    protected void setPerceptionList(PerceptionList p) 
    {
	knownPerceptions = p;    
    }
   
    /**
     * @param ac
     */
    protected void setPerceptionListPointer(MACSActionSelector ac) 
    {
	perceptionListPointer = ac;    
    }
   
    /**
     * @return MACS.PerceptionList
     */
    protected PerceptionList getPerceptionList() 
    {
	//return perceptionListPointer.getPerceptionList();
	return knownPerceptions;    
    }
   
    private void fillPossibleActionsArray() 
    {
	Action action = new Action(getActionForm());

	nbActions = (short)1;
	for (byte i=0 ; i<action.getLength() ; i++)
	    nbActions = (short)(nbActions * action.getNbSpecValues(i));
	possibleActionsArray = new Action[nbActions];

	for (short i=0 ; i<nbActions ; i++)
	    possibleActionsArray[i] = new Action(getActionForm());

	for (short i=0 ; i<nbActions ; i++)
	    {
		short div = 1;
		for (byte j=(byte)(action.getLength()-(byte)1) ; j>= 0 ; j--)
		    {
			possibleActionsArray[i].setSpecializedToken(j,(byte)(i/div));
			div = (short)(div*action.getNbSpecValues(j));
		    }
	    }
	if (DEBUG)
	    {
		System.out.println("possibleActionsArray : ");
		for (short i=0 ; i<nbActions ; i++)
		    {
			System.out.print(possibleActionsArray[i] + " ");
		    }
		System.out.println("");
	    }    
    }
   
    /**
     * @param input
     * @param action
     */
    public final void addNewClassifier(String input, String action) 
    {
	loader.addClassifier(input + " " + action);// cf. Constructeur...    
    }
   
    /**
     * @return short
     */
    protected final short getNbActions() 
    {
	return nbActions;    
    }
   
    /**
     * @param index
     * @return Classifier.Action
     */
    protected final Action getAction(short index) 
    {
	return possibleActionsArray[index];    
    }
   
    /**
     * @return Classifier.Action
     */
    public final Action getRandomAction() 
    {
	return getAction((short)(StaticRandom.nextDouble()*getNbActions()));    
    }
   
    /**
     * @param action
     * @return short
     */
    protected final short getIndexOf(Action action) 
    {
	short i = 0;
	while (i<nbActions)
	    {
		if (possibleActionsArray[i].equals(action))
		    return (i);
		i++;
	    }

	return ((short)(-1));    
    }
   
    /**
     * @return MACS.ClassifierSystemTraceManager
     */
    public final ClassifierSystemTraceManager getTraceManager() 
    {
	return traceManager;    
    }
   
    /**
     * @return MACS.GeneralizationManager
     */
    protected final GeneralizationManager getGeneralizationManager() 
    {
	return generalizationManager;    
    }
   
    /**
     * @return String
     */
    protected final String getPerceptionForm() 
    {
	return perceptionForm;    
    }
   
    /**
     * @return String
     */
    protected final String getActionForm() 
    {
	return actionForm;    
    }
   
    /**
     * @param trans
     * @param lastActive
     */
    protected final void updateModelOfTransitions(Transition trans, boolean lastActive) 
    {
	Action lastAction = trans.getLastAction();
	Perception lastPerception = trans.getLastPerception();
	if (lastAction!=null) traceManager.setLastAction(lastAction.toString());
	else traceManager.setLastAction("null");

	if (lastPerception!=null) traceManager.setLastPerception(lastPerception.toString());
	else traceManager.setLastPerception("null");
	
	if (lastActive && (lastPerception!=null) && (lastAction!=null))
	    {
		//System.out.println("1 : " + toString(true));
		if (ClassifierSystemParameters.conditionCovering)
		    conditionCoveringManager.run(lastPerception, lastAction, trans.getCurrentPerception());
		//		System.out.println("2 : " + toString(true));
		if (ClassifierSystemParameters.anticipationLearning)
		    anticipationLearningManager.run(lastPerception, lastAction, trans.getCurrentPerception());
		//System.out.println("3 : " + toString(true));
		if (ClassifierSystemParameters.mutspecDiscovering) 
		    specializationManager.run(lastPerception, lastAction);
		//System.out.println("4 : " + toString(true));
		if (ClassifierSystemParameters.generalization)
		    generalizationManager.run(lastPerception, lastAction, trans.getCurrentPerception());
		//System.out.println("5 : " + toString(true));
	    }    
    }
   
    /**
     * This method is used to measure the percentage of knowledge
     * 
     * @param classifierList
     * @return double
     */
    public final double checkForConformityWith(ClassifierSystem classifierList)
    {
	if (DEBUG) System.out.println("checkForConformityWith");
	int nbSuccess = 0;

	for (short i=0 ; i<classifierList.getNbClassifiers() ; i++)
	    {
		AnticipatoryClassifier actualTransition = (AnticipatoryClassifier)classifierList.getClassifier(i);
		Vector anticip = computeAnticipatedSituations(actualTransition.getCondition(),actualTransition.getAction());
		if (anticip.size()==1)
		    {
			Perception p = (Perception)((CouplePF)anticip.get(0)).getFirst();
			if (actualTransition.matchAnticipation(p))    nbSuccess++;
		    }
	    }
	if (DEBUG) System.out.println("fin de checkForConformityWith");
	return (double)((double)nbSuccess / (double)classifierList.getNbClassifiers());    
    }
   
    /**
     * @return String
     */
    public final String getStepInformation() 
    {
	return getTraceManager().getStepInformation(1);    
    }
   
   /**
    * This method returns the number of actions covered for a given perception by the ClassifierSystem
    * It is used to see if the agent should go again to the situation to improve its model of transitions
    * @param situation
    * @return byte : the number of actions covered
    */
    protected final byte nbActionsCovered(Perception situation) 
    {
	byte nbActionsCovered = 0;
		for (short j=0; j<getNbActions(); j++)
		    {
			boolean covered = false;
			for (short i=0; i<getNbClassifiers(); i++)
			    {
				if (getClassifier(i).matchCondition(situation) && getClassifier(i).getAction().equals(getAction(j)))
				    {
					covered = true;
					break;
				    }
			    }
			if (covered)    nbActionsCovered++;

		    }
	return nbActionsCovered;    
    }
   
    /**
     * renvoie la liste des anticipations concrtes possibles pour la perception 
     * courante
     * et l'action courante.
     * Vector  est un vecteur de triplets : 
     * [0] anticipations (Perceptions)
     * [1] rels (informations)
     * [2] booleen(reliabilities)
     * 
     * @param perception
     * @param action
     * @return java.util.Vector
     */
    public final Vector computeAnticipatedSituations(Perception perception, Action action) 
    {
	Vector effectparts = new Vector();
	Vector infoGains = new Vector();

	for (short i=0 ; i<getNbClassifiers() ; i++)
	    {
		AnticipatoryClassifier acl = (AnticipatoryClassifier)getClassifier(i);
		double level = acl.getEvaluationLevel(ClassifierSystemParameters.nbEvaluationsToSpecialize, ClassifierSystemParameters.nbFailuresToRemove, ClassifierSystemParameters.nbSuccessToProbe);
		Double info = new Double(level);
		if (acl.matches(perception,action))
		    {
			if (!effectparts.contains(acl.getAnticipation()))
			    {
				effectparts.add(acl.getAnticipation());
				infoGains.add(info);
			    }
			else 
			    {
				// si l'anticipation est y dj, mais avec un meilleur infoGain
				int anticipationIndex = effectparts.indexOf(acl.getAnticipation());
				if (((Double)infoGains.get(anticipationIndex)).doubleValue()< level)
				    {
					infoGains.set(anticipationIndex, info);
				    }
			    }
		    }
	    }
	// calcul de 1-infoGain (voir algo Pierre pour le terme exact)
	for (int i=0; i<infoGains.size() ; i++)
	    infoGains.set(i,new Double(1.0 - ((Double)infoGains.get(i)).doubleValue()));
	
	PerceptionList ps = getPerceptionList();
	/*
	  if (localDEBUG)//
	  { 
	  System.out.println("infoGains avant : " + infoGains);
	  System.out.print("effectparts avant : [");
	  for (int i=0; i<effectparts.size() ; i++)
	  System.out.println(effectparts.get(i));
	  System.out.println("]\nperceptions avant : " + ps);
	  }
	*/
	return Integrator.buildAnticipatedSituations(effectparts, infoGains, ps);    
    }
   
    /**
     * @param verboseForClassifiers
     * @return String
     */
    public final String toString(boolean verboseForClassifiers) 
    {
	return "--- List of Classifiers ---\n" + super.toString(verboseForClassifiers);    
    }
   
    /**
     * @return String
     */
    public final String toString() 
    {
	return (toString(false));    
    }
}
