
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import Utils.ParameterSet;

/**
 * This class is used to store all the parameters used in MACS.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
*/

public final class ClassifierSystemParameters 
{
   public static String systemName;

    public static int nbPlanningSteps;
    public static String plannerType;

    public static int nbCriteria;
   public static String[] criteriaNames;
   public static double[] criteriaDiscountFactor;
   public static double[] criteriaDefaultValue;
    public static double[] criteriaIncrLearningRate;
    public static double[] criteriaDecrLearningRate;
      
   public static boolean stateSearch = false;
   public static boolean infoSearch = false;
   public static boolean maximizeRewardFirst = false;

   public static double initialReliability;
   public static double learningRate;
   public static double eGreedy;
   public static byte nbFailuresToRemove;
   public static byte nbSuccessToProbe;
   public static byte nbEvaluationsToSpecialize;
   public static byte classifierMemorySize;
   public static boolean conditionCovering;
   public static boolean obsoleteConditionSelection;
   public static boolean anticipationLearning;
   
   public static boolean mutspecDiscovering;
   public static boolean generalization;
   public static boolean traceConditionCovering = false;
   public static boolean traceObsoleteConditionSelection = false;
   public static boolean traceAnticipationCovering = false;
   public static boolean traceAnticipationSelection = false;
   public static boolean traceMutspecCreation = false;
   public static boolean traceMutspecRemoval = false;
   public static boolean traceMutgenCreation = false;
   public static boolean traceMutgenRemoval = false;
   public static boolean traceGeneralInfos = false;
   public static boolean traceInputs = false;
   public static boolean traceOutput = false;
   public static boolean traceStatistics = false;
   public static short traceClassifierListFrequency;
   public static short traceSituationValuesFrequency;
   public static boolean traceTime = false;
   
   /**
    * @param fileName
    */
   protected static void init(String fileName) 
   {	
	ParameterSet parameters = new ParameterSet(fileName);

	nbPlanningSteps = Integer.valueOf(parameters.getValue("nbPlanningSteps")).intValue();
	nbCriteria = Integer.valueOf(parameters.getValue("nbCriteria")).intValue();
	criteriaNames = new String[nbCriteria];
	criteriaDiscountFactor = new double[nbCriteria];
	criteriaDefaultValue = new double[nbCriteria];
	criteriaIncrLearningRate = new double[nbCriteria];
	criteriaDecrLearningRate = new double[nbCriteria];
	for (int i=0;i<nbCriteria;i++)
	    {
		criteriaNames[i] = parameters.getValue("criteriaNames",i);
		criteriaDiscountFactor[i] = Double.valueOf(parameters.getValue("criteriaDiscountFactor",i)).doubleValue();
		criteriaIncrLearningRate[i] = Double.valueOf(parameters.getValue("criteriaIncrLearningRate",i)).doubleValue();
		criteriaDecrLearningRate[i] = Double.valueOf(parameters.getValue("criteriaDecrLearningRate",i)).doubleValue();
		criteriaDefaultValue[i] = Double.valueOf(parameters.getValue("criteriaDefaultValue",i)).doubleValue();
		if (criteriaNames[i].equals("stateSearch")) stateSearch = true;
		else if (criteriaNames[i].equals("info")) infoSearch = true;
	    }
	
	systemName	                      = parameters.getValue("systemName");
	plannerType                           = parameters.getValue("plannerType");

	initialReliability                    = Double.valueOf(parameters.getValue("initialReliability")).doubleValue();

	eGreedy                               = Double.valueOf(parameters.getValue("eGreedy")).doubleValue();

	learningRate                          = Double.valueOf(parameters.getValue("learningRate")).doubleValue();

	nbFailuresToRemove                    = Byte.valueOf(parameters.getValue("nbFailuresToRemove")).byteValue();
	nbSuccessToProbe                      = Byte.valueOf(parameters.getValue("nbSuccessToProbe")).byteValue();
	nbEvaluationsToSpecialize             = Byte.valueOf(parameters.getValue("nbEvaluationsToSpecialize")).byteValue();

	classifierMemorySize                  = Byte.valueOf(parameters.getValue("classifierMemorySize")).byteValue();

	conditionCovering                     = convertSwitch(parameters.getValue("conditionCovering"));
	obsoleteConditionSelection            = convertSwitch(parameters.getValue("obsoleteConditionSelection"));
	anticipationLearning                  = convertSwitch(parameters.getValue("anticipationLearning"));
	mutspecDiscovering                    = convertSwitch(parameters.getValue("mutspecDiscovering"));
	generalization                        = convertSwitch(parameters.getValue("generalization"));

	traceConditionCovering                = conditionCovering && convertSwitch(parameters.getValue("traceConditionCovering"));
	traceObsoleteConditionSelection       = obsoleteConditionSelection && convertSwitch(parameters.getValue("traceObsoleteConditionSelection"));

	traceAnticipationCovering             = anticipationLearning && convertSwitch(parameters.getValue("traceAnticipationCovering"));
	traceAnticipationSelection            = anticipationLearning && convertSwitch(parameters.getValue("traceAnticipationSelection"));

	traceMutspecCreation                  = mutspecDiscovering && convertSwitch(parameters.getValue("traceMutspecCreation"));
	traceMutspecRemoval                   = mutspecDiscovering && convertSwitch(parameters.getValue("traceMutspecRemoval"));

	traceMutgenCreation                   = generalization && convertSwitch(parameters.getValue("traceMutgenCreation"));
	traceMutgenRemoval                    = generalization && convertSwitch(parameters.getValue("traceMutgenRemoval"));

	traceGeneralInfos                     = convertSwitch(parameters.getValue("traceGeneralInfos"));
	traceInputs                           = convertSwitch(parameters.getValue("traceInputs"));
	traceOutput                           = convertSwitch(parameters.getValue("traceOutput"));

	traceTime                             = convertSwitch(parameters.getValue("traceTime"));

	traceClassifierListFrequency           = Short.valueOf(parameters.getValue("traceClassifierListFrequency")).shortValue();
	traceSituationValuesFrequency         = Short.valueOf(parameters.getValue("traceSituationValuesFrequency")).shortValue();

	traceStatistics                       = convertSwitch(parameters.getValue("traceStatistics"));    
   }
   
   /**
    * @param str
    * @return boolean
    */
   private static boolean convertSwitch(String str) 
   {
	return (str.equals("on"));    
   }
}
