
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.*;
import Utils.Misc;
import Classifier.AnticipatoryClassifier;

/**
 * This class is used to trace the behavior of the LCS.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
*/
public final class ClassifierSystemTraceManager 
{
   private String orderedPerceptions;
   private String coordsOfOrderedPerceptions;
   private String lastPerception;
   private String lastAction;
   private ClassifierSystem classifierSystem;
   private AnticipatoryClassifierList classifiersCreatedByAnticipationCovering;
   private AnticipatoryClassifierList classifiersCorrectedByDefiningAnticipation;
   private AnticipatoryClassifierList classifiersCreatedByPerceptionCovering;
   private AnticipatoryClassifierList classifiersCreatedByMutSpec;
   private AnticipatoryClassifierList classifiersRemovedByAnticipationSelection;
   private AnticipatoryClassifierList classifiersRemovedByInutilityIdentification;
   private AnticipatoryClassifierList classifiersRemovedByMutSpec;
   private AnticipatoryClassifierList classifiersCreatedByGeneralization;
   private AnticipatoryClassifierList classifiersRemovedByGeneralization;
   
   /**
    * Warning: we only copy pointers.
    * @param cstm
    * @param cs
    */
   protected ClassifierSystemTraceManager(ClassifierSystemTraceManager cstm, ClassifierSystem cs) 
   {
	classifierSystem = cs;

	orderedPerceptions = cstm.orderedPerceptions;
	coordsOfOrderedPerceptions = cstm.coordsOfOrderedPerceptions;

	classifiersCreatedByAnticipationCovering = cstm.classifiersCreatedByAnticipationCovering;
	classifiersCorrectedByDefiningAnticipation = cstm.classifiersCorrectedByDefiningAnticipation;
	classifiersCreatedByPerceptionCovering = cstm.classifiersCreatedByMutSpec;
	classifiersCreatedByMutSpec = cstm.classifiersCreatedByPerceptionCovering;
	
	classifiersRemovedByAnticipationSelection = cstm.classifiersRemovedByAnticipationSelection;
	classifiersRemovedByInutilityIdentification = cstm.classifiersRemovedByInutilityIdentification;
	classifiersRemovedByMutSpec = cstm.classifiersRemovedByMutSpec;
	
	classifiersCreatedByGeneralization = cstm.classifiersCreatedByGeneralization;
	classifiersRemovedByGeneralization = cstm.classifiersRemovedByGeneralization;    
   }
   
   /**
    * @param cs
    */
   protected ClassifierSystemTraceManager(ClassifierSystem cs) 
   {
	classifierSystem = cs;    
   }
   
   /**
    * @return MACS.PerceptionList
    */
   protected PerceptionList getPerceptionList() 
   {
	return (PerceptionList)classifierSystem.getPerceptionList();    
   }
   
   protected final void setFree() 
   {
	classifierSystem = null;    
   }
   
   protected final void prepareForNewStep() 
   {
	classifiersCreatedByAnticipationCovering = new AnticipatoryClassifierList();
	classifiersCorrectedByDefiningAnticipation = new AnticipatoryClassifierList();
	classifiersCreatedByPerceptionCovering = new AnticipatoryClassifierList();
	classifiersCreatedByMutSpec = new AnticipatoryClassifierList();

	classifiersRemovedByAnticipationSelection = new AnticipatoryClassifierList();
	classifiersRemovedByInutilityIdentification = new AnticipatoryClassifierList();
	classifiersRemovedByMutSpec = new AnticipatoryClassifierList();

	classifiersCreatedByGeneralization = new AnticipatoryClassifierList();
	classifiersRemovedByGeneralization = new AnticipatoryClassifierList();

	lastPerception = "";
	lastAction = "";    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersCreatedByGeneralization() 
   {
	return classifiersCreatedByGeneralization;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersRemovedByGeneralization() 
   {
	return classifiersRemovedByGeneralization;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersCreatedByAnticipationCovering() 
   {
	return classifiersCreatedByAnticipationCovering;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersCorrectedByDefiningAnticipation() 
   {
	return classifiersCorrectedByDefiningAnticipation;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersCreatedByPerceptionCovering() 
   {
	return classifiersCreatedByPerceptionCovering;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersCreatedByMutSpec() 
   {
	return classifiersCreatedByMutSpec;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersRemovedByAnticipationSelection() 
   {
	return classifiersRemovedByAnticipationSelection;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersRemovedByInutilityIdentification() 
   {
	return classifiersRemovedByInutilityIdentification;    
   }
   
   /**
    * @return MACS.AnticipatoryClassifierList
    */
   protected final AnticipatoryClassifierList getClassifiersRemovedByMutSpec() 
   {
	return classifiersRemovedByMutSpec;    
   }
   
   /**
    * @param acl
    * @param val
    */
   public final void addClassifierRemovedByGeneralization(AnticipatoryClassifier acl, boolean val) 
   {
	classifiersRemovedByGeneralization.addClassifier(acl,val);    
   }
   
   /**
    * @param acl
    * @param val
    */
   public final void addClassifierCreatedByGeneralization(AnticipatoryClassifier acl, boolean val) 
   {
	classifiersCreatedByGeneralization.addClassifier(acl,val);    
   }
   
   /**
    * @return short
    */
   protected final short getNbClassifiers() 
   {
	return classifierSystem.getNbClassifiers();    
   }
   
   /**
    * @param orderedPerc
    */
   public final void setOrderedPerceptions(String orderedPerc) 
   {
	orderedPerceptions = orderedPerc;
	//	System.out.println(this + " : Ordered Perc => " + orderedPerceptions);
   }
   
   /**
    * @param coordsOfOrderedPerc
    */
   public final void setCoordsOfOrderedPerceptions(String coordsOfOrderedPerc) 
   {
	coordsOfOrderedPerceptions = coordsOfOrderedPerc;    
   }
   
   /**
    * @param la
    */
   public final void setLastAction(String la) 
   {
	lastAction = la;    
   }
   
   /**
    * @param lp
    */
   public final void setLastPerception(String lp) 
   {
	lastPerception = lp;    
   }
   
   /**
    * @param time
    * @return String
    */
   protected final String getStepInformation(long time) 
   {
	String str = new String();
	/*
	if (time >25) 
	    {
		ActionSelector.DEBUG = true;
		PayoffModel.localTraceFlag = true;
		Integrator.localDEBUG = true;
		ClassifierSystem.localDEBUG = true;
		System.out.println(classifierSystem.toString(true));
	    }
	*/
	if (ClassifierSystemParameters.traceConditionCovering
	    && (classifiersCreatedByPerceptionCovering.getNbClassifiers()!=0))
	    str = str + "\n"
		+ Misc.putSideBySide("Cond Covering :",
				     classifiersCreatedByPerceptionCovering.toString(false), (short)20);

	if (ClassifierSystemParameters.obsoleteConditionSelection && classifiersRemovedByInutilityIdentification.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("Obsolete :",
				     classifiersRemovedByInutilityIdentification.toString(false), (short)20);

	if (ClassifierSystemParameters.traceAnticipationCovering && classifiersCorrectedByDefiningAnticipation.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("Ant Defining :",
				     classifiersCorrectedByDefiningAnticipation.toString(false), (short)20);

	if (ClassifierSystemParameters.traceAnticipationCovering && classifiersCreatedByAnticipationCovering.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("Ant Covering :",
				     classifiersCreatedByAnticipationCovering.toString(false), (short)20);

	if (ClassifierSystemParameters.traceAnticipationSelection && classifiersRemovedByAnticipationSelection.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("Ant Selection :",
				     classifiersRemovedByAnticipationSelection.toString(false), (short)20);

	if (ClassifierSystemParameters.traceMutspecCreation && classifiersCreatedByMutSpec.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("MutSpec Creation :",
				     classifiersCreatedByMutSpec.toString(false), (short)20);

	if (ClassifierSystemParameters.traceMutspecRemoval && classifiersRemovedByMutSpec.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("MutSpec Removal :",
				     classifiersRemovedByMutSpec.toString(false), (short)20);

	if (ClassifierSystemParameters.traceMutgenCreation && classifiersCreatedByGeneralization.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("MutGen Creation :",
				     classifiersCreatedByGeneralization.toString(false), (short)20);

	if (ClassifierSystemParameters.traceMutgenRemoval && classifiersRemovedByGeneralization.getNbClassifiers()!=0)
	    str = str + "\n"
		+ Misc.putSideBySide("MutGen Removal :",
				     classifiersRemovedByGeneralization.toString(false), (short)20);

	if (ClassifierSystemParameters.traceGeneralInfos)
	    str = str + "\n"
		+ Misc.putSideBySide("General Infos :",
				     new String (
				       "Last Perception  " + lastPerception + "\n"
				     + "Last Action      " + lastAction + "\n"
                                     + "Nb Classifiers   " + classifierSystem.getNbClassifiers() + "\n")
				     + "Nb Goods :       " + classifierSystem.getNbWellAnticipatingClassifiers() + " \n"
				     + "Nb Perceptions : " + getPerceptionList().size() + " \n"
				     ,(short)20 );

	//System.out.println(time);

	if ( (ClassifierSystemParameters.traceClassifierListFrequency!=0)
	     && ((time%ClassifierSystemParameters.traceClassifierListFrequency)==0)
	     && (classifierSystem.getNbClassifiers()!=0) )
	    {
	    str = str + "\n"
		+ Misc.putSideBySide("Resulting Set :",
				     classifierSystem.toString(true), (short)20);
	    }
	if ( (ClassifierSystemParameters.traceSituationValuesFrequency!=0)
	     && ((time%ClassifierSystemParameters.traceSituationValuesFrequency)==0)
	     && (getPerceptionList().size()!=0) )
	    {
		//	System.out.println(this + " : X Ordered Perc => " + orderedPerceptions);

	    str = str + "\n"
		+ Misc.putSideBySide("Situation Values :",
				     getPerceptionList().toString(orderedPerceptions,
								  coordsOfOrderedPerceptions),
				     (short)20);
	    }
	return str;
   }
}
