
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.Vector;
import java.lang.reflect.Array;
import Utils.StaticRandom;
import Classifier.Action;
import Classifier.Perception;
import Classifier.Anticipation;
import Classifier.MACSAnticipation;
import Classifier.YACSAnticipation;
import Classifier.AnticipatoryClassifier;
import Classifier.Token;

/**
 * The ConditionCoveringManager class is in charge of adding new classifiers when 
 * no one matches the last situation experienced.
 * The difficulty consists in setting the added classifier to the good level of 
 * generality.
 * The only method called from outside is run().
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
 */
final class ConditionCoveringManager 
{
    public static boolean DEBUG = false;//true;//

   private ClassifierSystem classifierSystem;
   
   /**
    * @param cs
    */
   protected ConditionCoveringManager(ClassifierSystem cs) 
   {
	classifierSystem = cs;    
   }
   
   /**
    * This method scatters the model of Transitions into subsets
    * For each classifier, if the classifier matches the current perception/action 
    * couple, it is put into a set whose index corresponds to its first specialized symbol.
    * This method should disappear : MACS should make one model per attribute,
    * so that no scattering is necessary...
    * 
    * @param lastAction
    * @param currentPerception
    * @return MACS.AnticipatoryClassifierList[]
    */
   private AnticipatoryClassifierList[] splitClassifierSystemIntoConcernedSubsets(Action lastAction, Perception currentPerception) 
   {
	AnticipatoryClassifierList[] subSets = new AnticipatoryClassifierList[currentPerception.getLength()];
	for (short i=0; i<currentPerception.getLength() ; i++) subSets[i] = new AnticipatoryClassifierList();
	
	for (short i=0; i<classifierSystem.getNbClassifiers() ; i++)
	{
	    AnticipatoryClassifier classifier = (AnticipatoryClassifier)classifierSystem.getClassifier(i);
	    Anticipation a = classifier.getAnticipation();
	    
	    if ((classifier.getAction().equals(lastAction)) && (a.matches(currentPerception)))
		{
		    short index = a.getIndexOfFirstSpecializedToken();
		    subSets[index].addClassifier(classifier, false);
		}		
	}	
	return subSets;
   }
   
   /**
    * This method checks whether the given Perception is matched by the condition part
    * of any member of the classifierList, given that this list contains classifier
    * whose action and effect part match the last action and the current situation.
    * @param classifierList the list among which we look for a matcher
    * @param perception the perception to be matched by the group
    * @return boolean true if one matches.
    */
   private boolean isTakenIntoAccount(AnticipatoryClassifierList classifierList, Perception perception) 
   {
	for (short i=0 ; i<classifierList.getNbClassifiers() ; i++)
	    if (classifierList.getClassifier(i).matchCondition(perception))	return true;
	return false;    
   }
   
   /**
    * This method generates recursively a list of condition parts more general than those of the list it receives,
    * checking that no condition in the list is in conflict with those of the classifierList
    * @param L1 the initial list of conditions to be genralized
    * @param L2 a list of more general conditions than those in L1
    * @param classifierList the list of similar classifiers matching the action and the effect, but not the last perception
    * @return MACS.PerceptionList
    */
   private static PerceptionList generalizeByLevels(PerceptionList L1, PerceptionList L2, AnticipatoryClassifierList classifierList, int level) 
   {
	if (DEBUG) System.out.println("generalizeByLevels level : " + level + " L1 : " + L1 + "\n L2 : " + L2 + "\n classifierList :" + classifierList);
	PerceptionList L3 = new PerceptionList();

	for (short i=0 ;i<L2.getNbPerceptions() ; i++)
	    {
		Perception p = L2.getPerception(i);
		if (!classifierList.containsClassifierMatching(p)) L3.add(p);
	    }
	
	if (DEBUG)	System.out.println("reste L3 : " + L3);
	if (L3.getNbPerceptions()==0)	    return L1;
	else
	    {
		// patch osd 26/09/03
		// incomplete heuristic so as to prevent an explosion 
		// instead of 
		return generalizeByLevels(L3, L3.buildMoreGeneralPerceptionList(), classifierList, level+1);    
		//PerceptionList L4 = new PerceptionList();
		//L4.add(L3.getRandomPerception());
		//return generalizeByLevels(L4, L4.buildMoreGeneralPerceptionList(), classifierList, level+1);    
	    }
   }
   
   /**
    * This method generalizes a condition part as much as possible to build a covering classifier.
    * The process still needs to be understood.
    * @param condition a tentative condition part matching the last perception that should be generalized further
    * @param classiferList the list of similar classifiers matching the action and the effect, but not the last perception
    * @return Classifier.Perception one of the more general condition parts needed to build a new classifier
    */
   private Perception generalizeEvenMore(Perception condition, AnticipatoryClassifierList classifierList) 
   {
	if (DEBUG) System.out.println("entre dans generalizeEvenMore : " + condition + "\n classifierList :" + classifierList);
	PerceptionList conditionList = new PerceptionList();
	conditionList.add(condition);
	PerceptionList p2 = PerceptionList.buildMoreGeneralPerceptionList(condition);
	conditionList = generalizeByLevels(conditionList, p2, classifierList, 0);
	if (DEBUG) System.out.println("sortie de generalizeEvenMore : " + conditionList);
	return (conditionList.getRandomPerception());    
   }
   
   /**
    * This method builds a general condition part for a new classifier that will be added
    * in the system to cover the last transition.
    * From each classifier of the classifierList, we build a new condition part which matches
    * the last perception, and we generalize it as much as possible.
    * Among all such conditions, we keep all that are  maximally general,
    * and we keep one at random among them.
    * This one is generalized further before being returned.
    * @param lastPerception the perception to be covered
    * @param classifierList the list of similar classifiers matching the action and the effect, but not the last perception
    * @return Classifier.Perception the general condition part needed to build a new classifier
    */
   private Perception buildCoveringCondition(Perception lastPerception, AnticipatoryClassifierList classifierList) 
   {
	if (classifierList.getNbClassifiers()==0) return new Perception(lastPerception.getForm());

	double minSpec = Double.POSITIVE_INFINITY;
	PerceptionList matchingList = new PerceptionList();
	for (short i=0 ; i<classifierList.getNbClassifiers() ; i++)
	    {
		Perception condition = new Perception(classifierList.getClassifier(i).getCondition());
		if (DEBUG) System.out.println("buildCoveringCondition :" + condition + " to match " + lastPerception);
		condition.correctSoAsToMatch(lastPerception);
		if (DEBUG) System.out.println("buildCoveringCondition after correct :" + condition);
		condition = generalizeEvenMore(condition, classifierList);
		if (DEBUG) System.out.println("buildCoveringCondition after generalize :" + condition);
		double spec = condition.getSpecialization();

		if (spec<minSpec)    matchingList = new PerceptionList();
		matchingList.add(condition);
		minSpec = spec;
	    }
	return generalizeEvenMore(matchingList.getRandomPerception(), classifierList);    
   }
   
   /**
    * The main method of the condition covering process.
    * First, the classifiers matching the last action and whose effect part matches the current perception
    * is split into sets corresponding to the specific attributes in the effect part.
    * Then, in each such set, we check if the lastPerception is matched.
    * If it is not, we build a new classifier matching this transition,
    * whose condition part is as general as possible.
    * This last step is acheived by buildCoveringCondition() (see above).
    * @param lastPerception the previous perception of the agent
    * @param lastAction the previous action of the agent
    * @param currentPerception the current perception of the agent
    */
   protected void run(Perception lastPerception, Action lastAction, Perception currentPerception) 
   {
	AnticipatoryClassifierList subLists[] = splitClassifierSystemIntoConcernedSubsets(lastAction, currentPerception);
	
	for (byte i=0; i<Array.getLength(subLists) ; i++)
	    {
		if (!isTakenIntoAccount(subLists[i], lastPerception))
		    {
			AnticipatoryClassifier newClassifier = new AnticipatoryClassifier(ClassifierSystemParameters.systemName,currentPerception.getForm(), lastAction.getForm(),ClassifierSystemParameters.initialReliability);

			if (DEBUG) System.out.println("avant buildCoveringCondition : OK");
			newClassifier.setCondition(buildCoveringCondition(lastPerception, subLists[i]));
			newClassifier.setAction(new Action(lastAction));
			if (DEBUG) System.out.println("aprs buildCoveringCondition : OK");

			Anticipation anticipation = null;

			if (ClassifierSystemParameters.systemName.equals("MACS"))
			    anticipation = new MACSAnticipation(currentPerception.getForm());
			else if (ClassifierSystemParameters.systemName.equals("YACS"))
			    anticipation = new YACSAnticipation(currentPerception.getForm());
			else
			{
			    System.out.println("ConditionCoveringManager: systemName not known");
			    System.exit(-1);
			}

			((Perception)anticipation).setToken(i,currentPerception.getToken(i));
			newClassifier.setAnticipation(anticipation);


			classifierSystem.addClassifier(newClassifier, false);
			if (ClassifierSystemParameters.traceConditionCovering) classifierSystem.getTraceManager().getClassifiersCreatedByPerceptionCovering().addClassifier(newClassifier,false);
		    }
		if (DEBUG) System.out.println("CoveringCondition i :" + i);
	    }
   }
}
