
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/
package MACS;

import java.util.*;
import Utils.StaticRandom;
import Classifier.Perception;
import Classifier.Action;
import Classifier.AnticipatoryClassifier;

/**
 * The GeneralizationManager class is in charge of generalizing classifiers in the 
 * model of transitions.
 * The only method called from outside is run().
 * The process can be described into for main stages:
 * - First, the system builds Set A of classifiers matching the last action and 
 * the current perception.
 * - Second, if all the classifiers in Set A anticipate correctly and have been 
 * evaluated enough,
 * the system builds a Set B of general classifiers from the Set A.
 * This is done in the "buildGeneralClassifierList" method.
 * - Third, the system builds from each set B a Set C of classifiers... see the papers!
 * - Fourth, the system builds from each set C a Set D of classifiers... see the papers!
 * Note that if we had implemented MACS as building one module per attribute,
 * this component would be far more efficient.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */

final class GeneralizationManager 
{
    public static boolean DEBUG = false;//true;//
   private ClassifierSystem classifierSystem;
   private AnticipatoryClassifierList specClassifiers;
   private AnticipatoryClassifierList genClassifiers;
   
   /**
    * @param cs
    */
   protected GeneralizationManager(ClassifierSystem cs) 
   {
	classifierSystem = cs;
	specClassifiers = new AnticipatoryClassifierList();
	genClassifiers = new AnticipatoryClassifierList();    
   }
   
   /**
    * Returns the list of already seen perceptions
    * 
    * @return MACS.PerceptionList
    */
   private PerceptionList getPerceptionList() 
   {
	return classifierSystem.getPerceptionList();    
   }
   
   /**
    * @param general
    * @param initial
    * @param incompatible
    * @return boolean
    */
   private boolean chooseBetweenGeneralAndInitial(AnticipatoryClassifierList general, AnticipatoryClassifierList initial, AnticipatoryClassifierList incompatible) 
   {
	boolean change = false;
	for (short i=0 ; i<general.getNbClassifiers() ; i++)
	    {
		AnticipatoryClassifier classifier = (AnticipatoryClassifier)general.getClassifier(i);
		AnticipatoryClassifierList matching = incompatible.selectClassifiersMatching(classifier.getCondition(), classifier.getAction());
		if ((matching.getNbClassifiers()!=0) && getPerceptionList().coversAtLeastOnePerceptionCoveredBy(matching, classifier))
		    general.setClassifier(i, initial.getClassifier(i));
		else
		    change = true;
	    }
	return change;    
   }
   
   /**
    * @param classifier1
    * @param classifier2
    * @return boolean
    */
   private boolean moreGeneralThan(AnticipatoryClassifier classifier1, AnticipatoryClassifier classifier2) 
   {
	if (!classifier1.matchCondition(classifier2.getCondition()))   		return false;
	if (classifier1.getCondition().isEqualOrMoreGeneralThan(classifier2.getCondition()))    return true;

	Iterator it = getPerceptionList().iterator();
	while (it.hasNext())
	    {
		Perception perception = (Perception)it.next();
		if (classifier2.matchCondition(perception) && !classifier1.matchCondition(perception)) return false;
	    }
	return true;    
   }
   
   /**
    * @param cs
    * @return boolean
    */
   private boolean optimizeClassifierList(AnticipatoryClassifierList cs) 
   {
	boolean change = false;
	for (short i=0 ; i<cs.getNbClassifiers() ; i++)
	    for (short j=0 ; j<cs.getNbClassifiers() ; j++)
		{
		    if (j!=i)
			{
			    AnticipatoryClassifier csfi = (AnticipatoryClassifier)cs.getClassifier(i);
			    AnticipatoryClassifier csfj = (AnticipatoryClassifier)cs.getClassifier(j);
			    if (moreGeneralThan(csfi, csfj))
				{
				    change = true;
				    cs.removeClassifier(j);
				    j--;
				}
			    else if (moreGeneralThan(csfj, csfi))
				{
				    change = true;
				    cs.removeClassifier(i);
				    j=0;
				}
			}
		}
	return change;    
   }
   
   /**
    * @param concernedClassifierList
    * @return MACS.AnticipatoryClassifierList
    */
   private AnticipatoryClassifierList buildGeneralClassifierList(AnticipatoryClassifierList concernedClassifierList) 
   {
	AnticipatoryClassifierList newGeneralClassifierList = new AnticipatoryClassifierList();

	for (short i=0; i<concernedClassifierList.getNbClassifiers(); i++)
	    {
		AnticipatoryClassifier acs = (AnticipatoryClassifier)concernedClassifierList.getClassifier(i);
		AnticipatoryClassifier newClassifier = acs.mutGenOne(ClassifierSystemParameters.initialReliability);
		if (DEBUG) System.out.println(acs + "\n -mutGenOne-> :" + newClassifier);
		newGeneralClassifierList.addClassifier(newClassifier, false);
	    }
	return newGeneralClassifierList;    
   }
   
   /**
    * From note 1, it seems that ClassifierSystem can contain several times the same 
    * classifier
    * @param concernedSet
    * @param newSet
    */
   private void updateClassifierSystem(AnticipatoryClassifierList concernedSet, AnticipatoryClassifierList newSet) 
   {
	for (short i=0 ; i<concernedSet.getNbClassifiers() ; i++)
	    classifierSystem.removeClassifier(concernedSet.getClassifier(i));
	for (short i=0 ; i<newSet.getNbClassifiers() ; i++)
	    classifierSystem.addClassifier(((AnticipatoryClassifier)newSet.getClassifier(i)),false);//note 1
	
	for (short i=0; i<concernedSet.getNbClassifiers(); i++)
	    {
		if ( (!newSet.contains(((AnticipatoryClassifier)concernedSet.getClassifier(i)), false) 
		      && (ClassifierSystemParameters.traceMutgenRemoval)) )
		    {
			classifierSystem.getTraceManager().addClassifierRemovedByGeneralization(((AnticipatoryClassifier)concernedSet.getClassifier(i)), false);
		    }
	    }
	for (short i=0; i<newSet.getNbClassifiers(); i++)
	    {
		if ( (!concernedSet.contains(((AnticipatoryClassifier)newSet.getClassifier(i)), false) 
		      && (ClassifierSystemParameters.traceMutgenCreation)) )
		    {
			classifierSystem.getTraceManager().addClassifierCreatedByGeneralization(((AnticipatoryClassifier)newSet.getClassifier(i)), false);
		    }
	    }    
   }
   
   /**
    * @param lastPerception
    * @param lastAction
    * @param currentPerception
    */
   private final void updateReliabilities(Perception lastPerception, Action lastAction, Perception currentPerception) 
   {
	for (short i=0 ; i<classifierSystem.getNbClassifiers() ; i++)
	    {
		AnticipatoryClassifier classifier = (AnticipatoryClassifier)classifierSystem.getClassifier(i);
		if (classifier.matchAction(lastAction))
		    {
			classifier.updateGeneralReliabilities(lastPerception, currentPerception,ClassifierSystemParameters.learningRate);
		    }
	    }    
   }
   
   /**
    * First, the system builds Set B of general classifiers from set A
    * If some improvement results, then the improvements are optimized and the model 
    * of transition is updated
    * 
    * @param concernedClassifierList
    * @param incompatibleClassifierList
    */
   private final void generalizeClassifiers(AnticipatoryClassifierList concernedClassifierList, AnticipatoryClassifierList incompatibleClassifierList) 
   {
	AnticipatoryClassifierList newGeneralClassifierList = buildGeneralClassifierList(concernedClassifierList);
	if (newGeneralClassifierList.getNbClassifiers()!=0)
	    {
		boolean change = chooseBetweenGeneralAndInitial(newGeneralClassifierList, concernedClassifierList, incompatibleClassifierList);
		if (DEBUG) System.out.println("change1 : OK" + newGeneralClassifierList);
		if ((change) && (newGeneralClassifierList.getNbClassifiers()>1))
		    {
			change = optimizeClassifierList(newGeneralClassifierList);
			
			if (change)
			    {
				updateClassifierSystem(concernedClassifierList, newGeneralClassifierList);
			    }
		    }
	    }    
   }
   
   /**
    * First, the system builds Set A of classifiers matching the last action and the 
    * current perception.
    * If all the classifiers in Set A anticipate correctly and have been evaluated 
    * enough,
    * then this set of classifier can be generalized
    * 
    * @param lastPerception
    * @param lastAction
    * @param currentPerception
    */
   protected final void run(Perception lastPerception, Action lastAction, Perception currentPerception) 
   {

	Vector concernedSets = classifierSystem.buildAnticipatoryConcernedSets(lastAction, currentPerception);
	Vector incompatibleSets = classifierSystem.buildIncompatibleSets(concernedSets, lastAction);
	if (DEBUG) 
	    {
		//System.out.println("concernedSets :" + concernedSets);
		System.out.println("incompatibleSets :" + incompatibleSets);
	    }
	for (int i=0 ; i<concernedSets.size() ; i++)
	    {
		AnticipatoryClassifierList acs = (AnticipatoryClassifierList)concernedSets.get(i);
	    if (acs.getNbClassifiers()!=0)
		{
		    updateReliabilities(lastPerception, lastAction, currentPerception);
		    if (acs.hasBeenEvaluatedEnoughForGeneralization())
			{
			    if (DEBUG) System.out.println("concernedSets(i) :" + acs);
			    generalizeClassifiers((AnticipatoryClassifierList)concernedSets.get(i), (AnticipatoryClassifierList)incompatibleSets.get(i));
			}
		}
	    }    
   }
}
