
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.Vector;
import java.util.Set;
import java.util.Iterator;

import Utils.StaticRandom;
import Classifier.Token;
import Classifier.Perception;

/**
 * The Integrator class is in charge of merging the partial anticipations of a 
 * list of classifiers
 * so as to predict the situations that the agent may reach at the next time step.
 * This list of anticipated situations is used by the ActionSelector to compute
 * the value of each possible next action.
 * The only method called from outside is "buildAnticipatedSituations"
 * For a given action in a given situation,
 * it receives a list of effectparts, 
 * the evaluation level of the classifier containing this effect part,
 * and the list of perceptions already seen in the environment.
 * From that, it computes the list of possible perceptions matching with
 * the effects parts.
 * The algorithm is as follows:
 * From the list of anticipations (e.g. [[?|?|?|0|?|?|?|?], [?|?|?|?|?|?|?|1], 
 * [?|?|?|?|?|?|?|0])
 * it computes for each attribute what are the possible values anticipated.
 * Then it filters the list of perceptions already seen in the environment:
 * For each of these perceptions, and for each attribute, if the value of this
 * attribute in this perception is not possible, then discard it.
 * Note that, though MACS only predicts one attribute per classifier 
 * (e.g. [?|?|?|?|?|?|?|0]),
 * the process will also work properly with classifiers predicting several 
 * attributes (e.g. [?|1|?|?|0|?|?|0]).
 * This has not been checked, but I'm confident about it! (Olivier Sigaud, 
 * 20/08/03)
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 **/
class Integrator 
{
    public static boolean DEBUG = false;//true;//
   
    public static boolean localDEBUG = false;//true;//
   
    /**
     * This method builds the structure containing each possible value for each token
     * The key is a token, the value is the information gain 
     * 
     * @param effectparts
     * @param infoGains
     * @return MACS.PossibleTokenMap[]
     */
    private static PossibleTokenMap[] buildTokPossibilitiesAndTokInformation(Vector effectparts, Vector infoGains) 
    {
    	int perceptionLength = ((Perception)effectparts.get(0)).getLength();
	PossibleTokenMap[] tokPossibilities = new PossibleTokenMap[perceptionLength];
	
	for (int i=0 ; i<perceptionLength ; i++)
	    {
		tokPossibilities[i] = new PossibleTokenMap();
	    }
	for (int i=0 ; i<effectparts.size() ; i++)
	    {
		Perception p = (Perception)(effectparts.get(i));
		for (byte j=0 ; j<perceptionLength ; j++)
		    {
			if (!p.isDontKnow(j))
			    {
				if (!tokPossibilities[j].containsKey(p.getToken(j)))
				    {
					tokPossibilities[j].add(p.getToken(j),((Double)infoGains.get(i)).doubleValue());
				    }
			    }
		    }
	    }
	return tokPossibilities;
    }
   
    /**
     * This method checks whether a perception from the list of perceptions already 
     * seen in the environment
     * is compatible with the list of tokens just created, i.e. if it can be 
     * anticipated from the model
     * of transitions
     * 
     * @param p
     * @param tokPossibilities
     * @return double
     */
    private static double isCompatibleWith(Perception p, PossibleTokenMap[] tokPossibilities) 
    {
	double retour = 1;
	for (byte i=0;i<tokPossibilities.length;i++)
	    {
		double val = isPossible(p.getToken(i),tokPossibilities[i]);
		if (val==-1) return -1;
		else retour = retour * val;
	    }
	return retour;    
    }
   
    /**
     * This method checks whether a given attribute
     * is compatible in the list of tokens just created
     * 
     * @param t
     * @param tokPossibilities
     * @return double
     */
    private static double isPossible(Token t, PossibleTokenMap tokPossibilities) 
    {
	if (tokPossibilities.size()>0)
	    {
		Set keyset  = tokPossibilities.keySet();
		Iterator it = keyset.iterator();
		while  (it.hasNext())
		    {
			Token tok = (Token)it.next();
			if (tok==t)
			    {
				Double info = (Double)tokPossibilities.get(tok);
				return info.doubleValue();
			    }
		    }
		return -1;
	    }
	else return 1;    
    }
   
    /**
     * private static Vector sort(Vector v)
     * {
     * TreeSet retour = new TreeSet(v);
     * Vector w = new Vector(retour);
     * return w;
     * }
     * 
     * @param tokPossibilities
     * @param referenceSet
     * @return java.util.Vector
     */
    private static Vector buildFilteredPerceptionList(PossibleTokenMap[] tokPossibilities, PerceptionList referenceSet) 
    {
	Vector realizedSet = new Vector();
	Iterator it = referenceSet.iterator();
	while  (it.hasNext())
	    {
		Perception anticip = (Perception)it.next();
		double value = isCompatibleWith(anticip,tokPossibilities);
		if (value!=-1)
		    {
			realizedSet.add(new CouplePF(anticip,1-value));
		    }
	    }
	return  realizedSet;    
    }
   
    /**
     * Main method: computes the list of anticipated situations (see the class comment)
     * 
     * @param effectparts
     * @param infoGains
     * @param perceptionList
     * @return Vector
     */
    public static final Vector buildAnticipatedSituations(Vector effectparts, Vector infoGains, PerceptionList perceptionList) 
    {
	if (effectparts.size()!=0)
	    {
		  if (DEBUG)
		  {
		  for (int i=0; i<infoGains.size() ; i++) 
		  System.out.println((Perception)effectparts.get(i) + " " + ((Double)infoGains.get(i)).doubleValue());
		  System.out.println(perceptionList);
		  }
		PossibleTokenMap[] possAndCert = buildTokPossibilitiesAndTokInformation(effectparts, infoGains);
		  if (DEBUG)
		  {
		  System.out.println(" possAndCert :  ");
		  for (int i=0; i<possAndCert.length ; i++)
		  {
		  System.out.println(possAndCert[i]);
		  }
		  }
		Vector filtered = buildFilteredPerceptionList(possAndCert, perceptionList);
		  if (localDEBUG)
		  {
		  System.out.println("retour -------->");
		  for (int i=0; i<filtered.size() ; i++) 
		  System.out.println(((Perception)((CouplePF)filtered.get(i)).getFirst()) + " " + ((CouplePF)filtered.get(i)).getSecond());
		  System.out.println("   ");
		  }
		return filtered;
	    }
	else 	if (DEBUG)	System.out.println("retour vide");
	return new Vector();    
    }
   
    /**
     * @param args[]
     */
    public static void main(String args[]) 
    {
	StaticRandom.setSeed(1);
	Vector ef = new Vector();
	Vector info = new Vector();
	PerceptionList ps = new PerceptionList();
	for (short i=0;i<50;i++)
	    {
		Perception p = new Perception("[2|2|2|2|2|2|2|2|2|2]",true);
		ef.add(new Perception(p));
		ps.add(new Perception(p));
		info.add(new Double(0.0));
	    }
	Integrator.buildAnticipatedSituations(ef,info,ps);    
    }
}
