
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.*;
import Utils.StaticRandom;
import Utils.ValuesDistribution;
import Classifier.Perception;
import Classifier.Action;
import Classifier.Transition;

/**
 * The MACSActionSelector class is in charge of choosing the next action
 * both in the context of actual action and in the context of
 * "mental" propagation of values.
 * In order to choose the best action, it must access both
 * the model of transitions and the models of payoff.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */
final class MACSActionSelector
{
    public static boolean DEBUG = false;//true;//

    protected ClassifierSystem modelOfTransitions;
    protected MultiModelOfPayoff modelOfPayoff;
    protected ModularClassifierSystemTraceManager traceManager;
    protected double maxValueOfPossibleActions;

    protected int nbPossibilities;
    private HierarchicalAgregator agregator;
   private ExploMemory exploMemory;

   protected Action actions[];
   protected Perception anticipations[];
    private	int modelNumber[];
    /**
     * @param tm
     * @param modTrans
     */
    protected MACSActionSelector(ModularClassifierSystemTraceManager tm, ClassifierSystem modTrans, MultiModelOfPayoff mp, ExploMemory explo)
    {
	modelOfPayoff = mp;
	traceManager = tm;
	modelOfTransitions = modTrans; 
	if (ClassifierSystemParameters.stateSearch)	exploMemory = explo;
	modelNumber = new int[ClassifierSystemParameters.nbCriteria];
	for (int i=0;i<ClassifierSystemParameters.nbCriteria;i++)
	    modelNumber[i] = i;
    }
    /**
     * @return MACS.ModelOfPayoff
     */

    public MultiModelOfPayoff getModelOfPayoff()
    {
	return modelOfPayoff;
    }
    /**
     * @return MACS.PerceptionList
     */
    public PerceptionList getPerceptionList() 
    {
	return modelOfPayoff.getPerceptionList();    
    }
   
    /**
     * Choose a random action
     * 
     * @return Classifier.Action
     */
    private Action selectActionAtRandom() 
    {
	return modelOfTransitions.getRandomAction();    
    }
    public void setModelNumber(int model, int value)
    {
	modelNumber[model] = value;
    }
   
    /**
     * Init all the tables
     * @param nbPoss : number of situations that are supposed to be reachable
     */
    protected final void initArrays(int nbPoss) 
    {
	agregator = new HierarchicalAgregator(nbPoss);
	nbPossibilities = nbPoss;
	actions = new Action[nbPossibilities];
	anticipations = new Perception[nbPossibilities];
    }
   
    /**
     * @param possib
     */
    protected final void fillArrays(Vector[] possib, Perception currentPerception) 
    {
	Vector vActions = possib[0];
	Vector vAnticipations = possib[1];
	Vector vInfoGains = possib[2];

	if (DEBUG)
	    {
		System.out.println("****************************************************************************");
		System.out.println("fillArrays : before :");
		System.out.println("vActions :" + vActions);
		//		System.out.println("vInfoGains :" + vInfoGains);
		for (int i=0 ; i<vAnticipations.size() ; i++)
		    {
			System.out.println("anticipations["+i+"] : " + ((Perception)vAnticipations.get(i)));
		    }
	    }
	for (int i=0 ; i<vAnticipations.size() ; i++)
	    {
		actions[i] = (Action)vActions.get(i);
		anticipations[i] = ((Perception)vAnticipations.get(i));
		if (anticipations[i]!=null)
		    {
			for (int j=0 ; j<ClassifierSystemParameters.nbCriteria; j++)
			    {
				agregator.setValue(j,i,(modelOfPayoff.getModel(j).getUpdatedExpectedValue(anticipations[i])));
				if (DEBUG)
				    {
					System.out.println(ClassifierSystemParameters.criteriaNames[j] + " : " + modelOfPayoff.getModel(j).getImmediateValue(anticipations[i]) + " ; " + modelOfPayoff.getModel(j).getExpectedValue(anticipations[i]));
				    }
			    }
		    }
		else
		    {
			for (int j=0 ; j<ClassifierSystemParameters.nbCriteria; j++)
			    {
				agregator.setValue(j,i,ClassifierSystemParameters.criteriaDefaultValue[j]);
			    }
		    }
	    }
	if (DEBUG)	System.out.println("fin de fillArrays");    
    }

    /**
     * This method returns the index of the best possible action
     * -> positive or null if there is one
     * -> -1 if there is no
     * -> -2 if the possibilities are equivalent
     * 
     * The algorithm is the following: for each action, retreive the best anticipated situation
     * corresponding to that action and its value.
     * Then pick the best anticipated situation among those generated by each action
     * (or one of them if several share the same value)
     * 
     * @param hierarchyLevel : actionSelection Mode : 0 = information, 1 = external reward, 2 = rehearsal
     * @param totallyGreedy : if true, choose the best action. If not, choose the worst
     * @return int : the number of the best anticipated situation
     *
     * Note: the number of the best anticipated situation is also the number of the corresponding action,
     * beacuse tables have been generated that way (@see fillArrays)
     *
     **/
    private int selectBestPossibility(int hierarchyLevel, boolean totallyGreedy)
    {
	double maxValue = Double.NEGATIVE_INFINITY;
	boolean indifferent = true;
	Vector indexesMax = new Vector();

	for (short i=0; i<modelOfTransitions.getNbActions(); i++)
	    {
		Action action = modelOfTransitions.getAction(i);
		if (DEBUG) System.out.println("Action " + action);
		/*
		if (DEBUG) System.out.println("Possibilities :" + nbPossibilities);
		*/
		// call of the new, clearer version
		int bestPossForAction = selectPossibility(hierarchyLevel,action,totallyGreedy);
		/*
		// call of the old version
		int bestPossForAction;		

		if (!totallyGreedy)	bestPossForAction = selectWorstPossibilityForAction(action, hierarchyLevel);
		else			bestPossForAction = selectBestPossibilityForAction(action, hierarchyLevel);
		*/
		/*
		if (DEBUG) System.out.println("Possibilities :" + nbPossibilities);
		*/
		if (DEBUG) System.out.println("bestPossForAction " + bestPossForAction);
		if (bestPossForAction!=-1)
		    {
			double value = agregator.getValue(hierarchyLevel, bestPossForAction);
			if (DEBUG) System.out.println("value " + value);

			if ((maxValue!=value) && (maxValue!= Double.NEGATIVE_INFINITY)) indifferent = false;
			if (maxValue<=value)
			    {
				if (maxValue<value)
				    {
					indexesMax.removeAllElements();
					maxValue = value;
				    }
				indexesMax.addElement(new Integer(bestPossForAction));
			    }
		    }
		else	indifferent = false;
	    }
	maxValueOfPossibleActions = maxValue;

	if (indifferent)	return -2;

	int chosenIndex=-1;
	if (indexesMax.size()!=0)
	    {
		chosenIndex = ((Integer)indexesMax.get(StaticRandom.nextInt(indexesMax.size()))).intValue();
		if (DEBUG) System.out.println("I choose the solution " + chosenIndex + "  (" + maxValue + ")");
	    }
	return chosenIndex;    
    }
   
    /**
     * Performing a mental action
     * nb : several selection modes can be performed in parallel
     * 
     * @param currentPerception
     */
    protected final void updateModels(Perception currentPerception) 
    {
	//if (DEBUG) System.out.println("updateModels : " + currentPerception);
	for (int i=0;i<ClassifierSystemParameters.nbCriteria; i++)
	    {
		int chosenPossibility = selectBestPossibility(i, true);
		if (chosenPossibility==(-2)) 
		    chosenPossibility = agregator.pickPossibWithValue(maxValueOfPossibleActions,i);
		double value = agregator.getValue(i,chosenPossibility);
		modelOfPayoff.getModel(i).setExpectedValue(currentPerception, value);
		//if (DEBUG) System.out.println("updateModels : " + currentPerception + ":" + i + " => "+value);

	    }
    }
      
    /**
     * This method is slightly less efficient than Pierre's
     * (calling selectWorstPossibilityForAction and selectBestPossibilityForAction), 
     * but far more clear...
     * The inefficiency comes from the fact that, if we only consider in the ValuesDistribution
     * the values such that actions[i].equals(action), we loose memory of indexes...
     *
     * It happens that the efficiency difference is very small...
     * @return int :  the index of the anticipated situation chosen
     */

    private int selectPossibility(int hierarchyLevel, Action action, boolean best)
    {
	boolean matches = false;
	ValuesDistribution distrib = new ValuesDistribution((short)nbPossibilities);
	for (int i=0; i<nbPossibilities ; i++)
	    {
		if (actions[i].equals(action))
		    {
			matches = true;
			distrib.add(agregator.getValue(hierarchyLevel,i));
		    }
		else 
		    {
			if (best) distrib.add(Double.NEGATIVE_INFINITY);
			else distrib.add(Double.POSITIVE_INFINITY);
		    }
	    }
	if (!matches) return -1;
	else 
	    {
		if (DEBUG) System.out.println(distrib);
		if (best)	return distrib.computeIndexMax();
		else		return distrib.computeIndexMin();
	    }
    }

    /**
     * Performing actual actions
     * nb : several selection modes can be performed in parallel
     * 
     * @param currentPerception
     * @return Classifier.Action
     */
    protected final Action chooseActualAction(Perception currentPerception) 
    {
	// patch osd pour tester l'explo pure
	if (ClassifierSystemParameters.maximizeRewardFirst)
	    {
		setModelNumber(0,1);
		setModelNumber(2,1);
	    }
	int chosenPossibility=-3;
	for (int i=0;i<ClassifierSystemParameters.nbCriteria; i++)
	    {
		chosenPossibility = selectBestPossibility(modelNumber[i], true);
		if (chosenPossibility!=(-2))
		    {
			traceManager.setHierarchyLvlDecision(modelNumber[i]);
			break;
		    }
	    }
	if (chosenPossibility==(-2))
	    {
		traceManager.setHierarchyLvlDecision(-1);
		chosenPossibility = agregator.pickPossibWithValue(maxValueOfPossibleActions,modelNumber[0]);
	    }
	if (DEBUG) System.out.println("chosenPossibility : " + chosenPossibility);
	// we also learn when we perform actual actions
	// Note that the choice is computed a second time, this is inefficient
	updateModels(currentPerception);
	Action chosenAction = actions[chosenPossibility];    
	if (ClassifierSystemParameters.stateSearch)
	    {
		if (!exploMemory.containsKey(currentPerception))
		    {
		       exploMemory.add(currentPerception,modelOfTransitions.getNbActions());
		    }
		exploMemory.addAction(currentPerception,chosenAction);
	    }
	return chosenAction;
    }
   
   
    /**
     * Choose the action to be performed
     * if actualAction = true, it is the action of the agent.
     * Otherwise, it is a "mental" action in the model
     * possib contains the list of possible actions and information gains
     * to choose the best
     * 
     * @param lastAction
     * @param currentPerception
     * @param actualAction
     * @return Classifier.Action
     */
    public Action selectAction(Action lastAction, Perception currentPerception, boolean actualAction, Vector[] possib)
    {
	if (DEBUG) System.out.println("selectAction : " + currentPerception);

	if (actualAction)
	    if (StaticRandom.nextDouble()<ClassifierSystemParameters.eGreedy) return selectActionAtRandom();

	Action chosenAction = lastAction;

	initArrays(possib[1].size());
	fillArrays(possib,currentPerception);

	if (nbPossibilities>0)
	    {
		if (actualAction)   chosenAction = chooseActualAction(currentPerception);
		else		    updateModels(currentPerception);
	    }
	return chosenAction;    
    }
    /**
     * @return String
     */
    public final String toString() 
    {
	return modelOfPayoff.getModel(0).keySet().toString();    
    }

    // end of MACSActionSelector

    // ----------------------------- deprecated code -----------------------------------------   
    /**
     * This method returns one of the worst actions according to the selection mode
     * or -1 if there is no
     * 
     * @param action
     * @param hierarchyLevel : actionSelection Mode
     * @return int

    private int selectWorstPossibilityForAction(Action action, int hierarchyLevel) 
    {
	double minValue = Double.POSITIVE_INFINITY;
	Vector indexesMin = new Vector();

	for (int i=0; i<nbPossibilities ; i++)
	    {
		double value = getValue(hierarchyLevel,i);
		if (actions[i].equals(action) && (minValue>=value))
		    {
			if (minValue>value)
			    {
				indexesMin.removeAllElements();
				minValue = value;
			    }
			indexesMin.addElement(new Integer(i));
		    }
	    }

	if (indexesMin.size()==0)    return (-1);
	else
	    {
		if (DEBUG) System.out.println(indexesMin);
		return ((Integer)indexesMin.get((int)(StaticRandom.nextDouble()*indexesMin.size()))).intValue();
	    }    
    }
   

     * This method returns one of the best actions according to the selection mode
     * or -1 if there is no
     * 
     * @param action
     * @param hierarchyLevel : actionSelection Mode
     * @return int

    private int selectBestPossibilityForAction(Action action, int hierarchyLevel) 
    {
	double maxValue = Double.NEGATIVE_INFINITY;
	Vector indexesMax = new Vector();

	for (int i=0; i<nbPossibilities ; i++)
	    {
		double value = getValue(hierarchyLevel,i);
		if (actions[i].equals(action) && (maxValue<=value))
		    {
			if (maxValue<value)
			    {
				indexesMax.removeAllElements();
				maxValue = value;
			    }
			indexesMax.addElement(new Integer(i));
		    }
	    }

	if (indexesMax.size()==0)   return (-1);
	else
	    {
		if (DEBUG) System.out.println(indexesMax);
		return ((Integer)indexesMax.get((int)(StaticRandom.nextDouble()*indexesMax.size()))).intValue();
	    }    
    }
     */
}
