
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.*;
import Utils.StaticRandom;
import Classifier.AnticipatoryClassifier;
import Classifier.Transition;
import Classifier.Perception;
import Classifier.Action;
/**
 * This class is the main class of MACS.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
*/
public final class ModularClassifierSystem implements RuleBasedSystem 
{
    public static boolean DEBUG = false;//true;//
   
    private String paramsFileName;
    private String actionForm;
    private String perceptionForm;
    private final MACSActionSelector actionSelector;
    protected final MultiModelOfPayoff modelOfPayoff;
    private final ModularClassifierSystemTraceManager traceManager;
    private final ClassifierSystem modelOfTransitions;
    private final Anticipator anticipator;
    private final InfoValueEstimator infoGainEstimator;
    protected final AbstractPlanner planner;
   private ExploMemory exploMemory;
    private Action lastAction = null;
    private Perception lastPerception = null;
   
    /**
     * @param rulesForm
     * @param initString
     * @param paramsFile
     */
    public ModularClassifierSystem(String rulesForm, String initString, String paramsFile) 
    {
	ClassifierSystemParameters.init(paramsFile);
	modelOfTransitions = new ClassifierSystem(rulesForm,initString);

	perceptionForm = new String(modelOfTransitions.getPerceptionForm());
	actionForm = new String(modelOfTransitions.getActionForm());

	traceManager = new ModularClassifierSystemTraceManager(this);
	modelOfPayoff = new MultiModelOfPayoff();  
	exploMemory = new ExploMemory();  
	actionSelector = new MACSActionSelector(traceManager,modelOfTransitions,modelOfPayoff,exploMemory);
	if (ClassifierSystemParameters.plannerType.equals("RandomPlanner"))
	    planner = new RandomPlanner(actionSelector,ClassifierSystemParameters.nbPlanningSteps);
	else
	    planner = new SystematicPlanner(actionSelector,ClassifierSystemParameters.nbPlanningSteps);
	//modelOfTransitions.setPerceptionListPointer(actionSelector);
	modelOfTransitions.setPerceptionList(actionSelector.getPerceptionList());

	infoGainEstimator = new InfoValueEstimator(modelOfTransitions);
	anticipator = new   Anticipator(modelOfTransitions);
    }
   
    /**
     * @return MACS.ClassifierSystem
     */
    protected final ClassifierSystem getModelOfTransitions() 
    {
	return modelOfTransitions;    
    }
   
    /**
     * @return MACS.ClassifierSystemTraceManager
     */
    public final ClassifierSystemTraceManager getTraceManager() 
    {
	return modelOfTransitions.getTraceManager();    
    }
   
    /**
     * @return MACS.ModularClassifierSystemTraceManager
     */
    public final ModularClassifierSystemTraceManager getModularTraceManager() 
    {
	return traceManager;    
    }
     
    /**
     * @param currentReward
     * @param situation
     * @param goalReached
     * @return String
     */
    public final String run(double currentReward, String situation, boolean goalReached) 
    {

	Perception currentPerception = new Perception(situation, perceptionForm); 
	traceManager.prepareForNewStep(currentReward, currentPerception);
	Transition trans = new Transition(lastPerception, lastAction, currentReward, currentPerception);

	//	System.out.println("avant updateModelOfTransitions");
	modelOfTransitions.updateModelOfTransitions(trans,true);
	//	System.out.println("avant updatePayoffModels");

	modelOfPayoff.updateListOfPerceptions(trans, goalReached);
	anticipator.clear();

	Vector possib[] = anticipator.getPossibilityTables(currentPerception);

	if (ClassifierSystemParameters.infoSearch)
	    {
		for (int j=0 ; j<ClassifierSystemParameters.nbCriteria; j++)
		    {
			if (ClassifierSystemParameters.criteriaNames[j].equals("info"))	
			    {	
				infoGainEstimator.fillImmediateInfoValue(possib);
				modelOfPayoff.getModel(j).setImmediateInfoValue(infoGainEstimator, possib[1]);
			    }
		    }
	    }

	modelOfPayoff.updateImmediateValues(trans, goalReached, exploMemory);

	if (modelOfPayoff.getModel(0).size()>0)
	    planner.updatePolicy(anticipator);    

	lastPerception = currentPerception;
		
	//System.out.println("avant selectAction");
	lastAction = actionSelector.selectAction(lastAction, currentPerception, true, possib);

	if (lastAction==null)	return null;
	else			return lastAction.toString();
    }
   
    public final void prepareForNewProblem() 
    {
	lastPerception = null;
	lastAction = null;    
    }
   
    /**
     *  voir
     * @param system
     * @return double
     */
    public final double checkForConformityWith(ClassifierSystem system) 
    {
	return modelOfTransitions.checkForConformityWith(system);    
    }
   
    /**
     * @return short
     */
    public final short getNbClassifiers()
    {
	return modelOfTransitions.getNbClassifiers();    
    }
   
    /**
     *  voir
     * @return String
     */
    public final String getStepInformation() 
    {
	return traceManager.getStepInformation();    
    }
   
    /**
     * @param v1
     * @return String
     */
    public final String toString(boolean v1) 
    {
	return modelOfTransitions.toString(v1);    
    }
   
    /**
     * @return String
     */
    public final String toString() 
    {
	return toString(true);    
    }
}
