
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.*;
import Utils.Misc;
import Classifier.AnticipatoryClassifier;
import Classifier.Perception;

/**
 * This class is used to trace the behavior of MACS.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
*/
final class ModularClassifierSystemTraceManager 
{
   private long time = - 1;
   private double reward;
   private int hierarchyLvlDecision;
   private ModularClassifierSystem modularClassifierSystem;
   private Perception perception;
   
   /**
    * @param cs
    */
   protected ModularClassifierSystemTraceManager(ModularClassifierSystem cs) 
   {
	modularClassifierSystem = cs;    
   }
   
   /**
    * @param rw
    * @param per
    */
   protected final void prepareForNewStep(double rw, Perception per) 
   {
	modularClassifierSystem.getModelOfTransitions().getTraceManager().prepareForNewStep();

	perception = per;
	reward = rw;
	time++;    
   }
   
   /**
    * @return int
    */
   protected final int getHierarchyLvlDecision() 
   {
	return hierarchyLvlDecision;    
   }
   
   /**
    * @param lvl
    */
   public final void setHierarchyLvlDecision(int lvl) 
   {
	hierarchyLvlDecision = lvl;    
   }
   
   /**
    * @return long
    */
   protected final long getTime() 
   {
	return time;    
   }
   
   /**
    * @return Classifier.Perception
    */
   protected final Perception getPerception() 
   {
	return perception;    
   }
   
   /**
    * @return double
    */
   protected final double getReward() 
   {
	return reward;    
   }
   
   /**
    * @return short
    */
   protected final short getNbClassifiers() 
   {
	return modularClassifierSystem.getNbClassifiers();    
   }
   
   /**
    * @return String
    */
   protected final String getStepInformation() 
   {
	String str = new String();
	//if (time >100)
	// {

	if (ClassifierSystemParameters.traceTime)
	    str = str + "==== TIME " + time + " ====" + "\t";

	if (ClassifierSystemParameters.traceInputs)
	    str = str + "\n"
		+ Misc.putSideBySide("Inputs :",
				     new String("Reward     " + reward +"\nPerception " + perception), (short)20);

	if (ClassifierSystemParameters.traceGeneralInfos)
	    str = str + "\n"
		+ Misc.putSideBySide("General Infos :",
				     new String (
                                       "Nb Classifiers : " + modularClassifierSystem.getNbClassifiers() + "\n")
				     + "Nb Goods :       " + getNbWellAnticipatingClassifiers() + " \n"
				     + "Decision Lvl :   " + getHierarchyLvlDecision() + " \n"
				     ,(short)20 );

	str = str + "\n************** Module ***************\n"
	    +modularClassifierSystem.getModelOfTransitions().getTraceManager().getStepInformation(time);
	str = str + "------------------------------------------\n";
	//}
	return str;    
   }
   
   /**
    * @return int
    */
   private int getNbWellAnticipatingClassifiers() 
   {
	int nb=0;
	for (short i=0 ; i<modularClassifierSystem.getModelOfTransitions().getNbClassifiers() ; i++)
	    if (((AnticipatoryClassifier)modularClassifierSystem.getModelOfTransitions().getClassifier(i)).doesAnticipateWell(ClassifierSystemParameters.nbSuccessToProbe))
		nb++;

	return nb;    
   }
}
