
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.io.*;
import java.util.*;
import java.lang.*;
import java.lang.reflect.*;
import Utils.StaticRandom;
import Classifier.Perception;
import Classifier.Transition;

/**
 * This class takes care of the three models of payoff used by MACS.
 * NB: This class does not need the model of transitions,
 * everything is in the ActionSelector...
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */
final class MultiModelOfPayoff
{
    private static final boolean DEBUG = false;//true;//

    private PayoffModel[] models;
    protected FinalStateModel fsModel;
    protected PerceptionList knownPerceptions;
    private int nbModels;
   
    public MultiModelOfPayoff() 
    {
	nbModels = ClassifierSystemParameters.nbCriteria;
	models = new PayoffModel[nbModels];

	for (int i=0;i<nbModels;i++)
	    {
		models[i] = new PayoffModel(ClassifierSystemParameters.criteriaNames[i],ClassifierSystemParameters.criteriaIncrLearningRate[i], ClassifierSystemParameters.criteriaDecrLearningRate[i],ClassifierSystemParameters.criteriaDiscountFactor[i]);
	    }
	knownPerceptions = new PerceptionList();
	fsModel = new FinalStateModel();    
    }
   
    /**
     * @return MACS.PerceptionList
     */
    public PerceptionList getPerceptionList() 
    {
	return knownPerceptions;    
    }
   
    /**
     * @return MACS.PayoffModel
     */
    public PayoffModel getModel(int index) 
    {
	return models[index];    
    }
   
    /**
     * @param p
     * @return boolean
     */
    public final boolean isFinal(Perception p) 
    {
	return fsModel.get(p);    
    }
   
    /**
     * @return MACS.FinalStateModel
     */
    public FinalStateModel getFsModel() 
    {
	return fsModel;    
    }
   
    /**
     * @param perception
     * @return boolean
     */
    public final boolean containsNotFinalMatching(Perception perception) 
    {
	Iterator it = getPerceptionList().iterator();
	while (it.hasNext())
	    {
		if (((Perception)it.next()).matches(perception) && !isFinal(perception))
		    return true;
	    }
	return false;    
    }
   
    /**
     * @return String
     */
    public final String toString() 
    {
	return getPerceptionList().toString();    
    }
   
    /**
     * add the new perception just received to the models of payoff
     * 
     * @param trans
     * @param lastStep
     */
    public final void updateListOfPerceptions(Transition trans, boolean lastStep)
    {
	Perception perception = trans.getCurrentPerception();
	if (models[0].containsKey(perception))
	    {
		fsModel.put(perception, lastStep);
	    }
	else
	    {
		for (int i=0;i<nbModels;i++)
		    {
			models[i].init(new Perception(perception));
		    }
		fsModel.put(new Perception(perception), lastStep);
		knownPerceptions.add(new Perception(perception));
	    }
    }
   
    /**
     * update the immediate values in the models of rewards
     * 
     * @param trans
     * @param lastStep
     * @param explo
     */
    public final void updateImmediateValues(Transition trans, boolean lastStep, ExploMemory explo) 
    {
	Perception perception = trans.getCurrentPerception();
	for (int i=0;i<nbModels;i++)
	    {
		if (ClassifierSystemParameters.criteriaNames[i].equals("external"))
		    {
			if (DEBUG) System.out.println("updateImmediateValues " +  perception + " => reward : " + trans.getCurrentReward());
			models[i].updateImmediateValue(perception, trans.getCurrentReward());
		    }
		else if (ClassifierSystemParameters.criteriaNames[i].equals("rehearsal"))
		    {
			models[i].updateImmediateRehearsal(perception,0.5);
		    }
		else if (ClassifierSystemParameters.criteriaNames[i].equals("stateSearch"))
		    {
			if (explo.containsKey(perception)) models[i].updateImmediateValue(perception, explo.getBooleanImmediateValue(perception));
		    }
	    }
    }
}
