
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.*;
import Utils.StaticRandom;
import Classifier.Perception;

/**
 * The PayoffModel class implements a list of immediate values and expected values
 * associated to each possible perception.
 * It is organized as a Map whose key is a Perception and whose value
 * is a couple of doubles : the immediate value and the expected value.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */
final class PayoffModel extends TreeMap 
{
   private String modeltype;
   private double discount_factor;
   private double incr_learning_rate;
   private double decr_learning_rate;
    public static boolean traceFlag = false;//true;//
   public static boolean localTraceFlag = false;//true;//
   
   /**
    * @param ilr
    * @param dlr
    * @param df
    */
   public PayoffModel(String mt, double ilr, double dlr, double df) 
   {
       modeltype = mt;
       incr_learning_rate = ilr;
       decr_learning_rate = dlr;
       discount_factor = df;    
   }
   
   /**
    * @param p
    * @return MACS.CoupleDoubles
    */
   public CoupleDoubles getCouple(Perception p) 
   {
	return (CoupleDoubles)get(p);    
   }
   
   /**
    * @param f
    */
   public void setTraceFlag(boolean f) 
   {
	traceFlag = f;    
   }
   
   /**
    * @param p
    */
   public void init(Perception p) 
   {
       if (traceFlag) System.out.println(modeltype + " : Appel de init : "+p + " "+super.toString());
	super.put(p,new CoupleDoubles(0.0));    
   }
   
   /**
    * @param p
    * @return double
    */
   public double getImmediateValue(Perception p) 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de getImmediateValue : " + p + " " + super.toString());
	return ((CoupleDoubles)get(p)).getFirst();    
   }
   
   /**
    * @param p
    * @param v
    */
   public void setImmediateValue(Perception p, double v) 
   {
       if (traceFlag) System.out.println(modeltype + " : Appel de setImmediateValue : " + p + " " + v + " " + super.toString());    
	super.put(p,new CoupleDoubles(v));
   }
   
   /**
    * @param p
    * @param v
    */
   public void put(Perception p, CoupleDoubles v) 
   {
	super.put(p,v);
	if (traceFlag) System.out.println(modeltype + " : Appel de put : " + p + " " + v + " " + super.toString());    
   }
   
   /**
    * @param p
    * @return double
    */
   public double getExpectedValue(Perception p) 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de getExpectedValue : " + p + " " + super.toString());
	return ((CoupleDoubles)get(p)).getSecond();    
   }
   
   /**
    * @param p
    * @param v
    */
   public void setExpectedValue(Perception p, double v) 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de setExpectedValue : " + p + " " + v + " " + super.toString());
	CoupleDoubles c = getCouple(p);
	c.setSecond(v);    
   }
   
   /**
    * widrow-hoff himself (c'est le cas o on augmente, pour l'external reward)
    * @param currentReward
    * @param currentPerception
    */
   public final void updateImmediateValue(Perception currentPerception, double currentReward) 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de updateImmediateValue : reward : " + currentReward + " " + currentPerception + " " + super.toString());
	CoupleDoubles couple = getCouple(currentPerception);
	double old = couple.getFirst();
	couple.setFirst(((1.0-incr_learning_rate)*old + incr_learning_rate*currentReward));    
   }
   
   /**
    * @param currentReward
    * @param currentPerception
    */
    public final double getUpdatedExpectedValue(Perception perception)
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de getUpdatedExpectedValue : "  + perception + " " + super.toString());
	double immediate = getImmediateValue(perception);
	//if (traceFlag) System.out.println("UpdateExpectedValue immediate : "  + perception + " : " + immediate);
	double updatedValue = discount_factor*getExpectedValue(perception)+immediate;
	return updatedValue;
   }
   
    /**
     * Update the rehearsal values of situations depending on whether they are visited or not
     * Stochastic version
    * @param currentPerception : the situation visited
    * @param proba : the probability to update
    */
    public final void updateImmediateRehearsal(Perception currentPerception, double proba) 
    {

	if (localTraceFlag) 	System.out.println(modeltype + " : Appel de updateImmediateRehearsal : " + currentPerception);
	//for (int i=0;i<ps.size();i++)
	Iterator it = keySet().iterator();

	while (it.hasNext())
	    {
		Perception p = (Perception)it.next();
		//Perception p = (Perception)ps.get(i);
		//if (localTraceFlag) 
		//   {
		//System.out.println("before : " + p + " = " + getImmediateValue(p) + " / " + getExpectedValue(p));
		//  }

		if (p.equals(currentPerception))
		    {
			updateImmediateValue(p,0.0);
		    }
		else
		    { 
			double a = StaticRandom.nextDouble();
			//if (localTraceFlag) System.out.println(a);
			if (a<0.1)
			    {
				double b = StaticRandom.nextDouble();
				//if (localTraceFlag)  System.out.println(b);
				// double b = 0.01;
				if (b<proba)
				    {
					updateImmediateValue(p,1.0);
				    }
			    }
		    }
		//if (localTraceFlag) 
		//    {
		//System.out.println("after : " + p + " = " + getImmediateValue(p) + " / " + getExpectedValue(p));
		//   }
	    }    
    }
   
    /**
     * Update the rehearsal values of situations depending on whether they are visited or not
     * Deterministic version
    * @param currentPerception : the situation visited
    */
    public final void updateImmediateRehearsal(Perception currentPerception) 
    {
	//if (localTraceFlag) 	System.out.println(modeltype + " : Appel de updateImmediateRehearsal : " + currentPerception);
	Iterator it = keySet().iterator();

	while (it.hasNext())
	    {
		Perception p = (Perception)it.next();

		if (p.equals(currentPerception))
		    {
			updateImmediateValue(p,0.0);
		    }
		else
		    { 
			updateImmediateValue(p,1.0);
			/* pour synchroniser les appels de random
			double a = StaticRandom.nextDouble();
			//if (localTraceFlag) System.out.println(a);
			if (a<0.1)
			    {
				double b = StaticRandom.nextDouble();
			    }
			*/
		    }
	    }    
    }
    /**
     * Update the information values of situations from the infoValueEstimator
    * @param infoEstimator : the infoValueEstimator
    */
    public final void   setImmediateInfoValue(InfoValueEstimator infoEstimator, Vector anticip)
    {
	for (int i=0 ; i<anticip.size() ; i++)
	    {
		Perception p = (Perception)anticip.get(i);
		if (p!=null)
		    {
			updateImmediateValue(p,infoEstimator.getImmediateInfoGains(i));
		    }
	    }
    }
   protected final void resetExpectedValues() 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de resetExpectedValues : "+super.toString());
	Iterator it = values().iterator();
	while (it.hasNext())
	    {
		((CoupleDoubles)it.next()).setSecond(0.0);
	    }    
   }
   
   protected final void resetImmediateValues() 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de resetImmediateValues : "+super.toString());
	Iterator it = values().iterator();
	while (it.hasNext())
	    {
		((CoupleDoubles)it.next()).setFirst(0.0);
	    }    
   }
   
   /**
    * Manifestement, on ne peut faire un reset sur les ExpectedValues qui si toutes 
    * les ImmediateValues sont nulles
    * 
    * @return boolean
    */
   protected final boolean isPossibleToResetExpectedValues() 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de isPossibleToResetExpectedValues : "+super.toString());
	Iterator it = values().iterator();
	while (it.hasNext())
	    {
		CoupleDoubles value = ((CoupleDoubles)it.next());
		if (value.getFirst()!=0.0) return false;
	    }
	return true;    
   }
   
   /**
    * @return double
    */
   protected final double getMaxExpectedValue() 
   {
	if (traceFlag) System.out.println(modeltype + " : Appel de getMaxExpectedValue : "+super.toString());
	Iterator it = values().iterator();
	double max = Double.NEGATIVE_INFINITY;
	while (it.hasNext())
	    {
		CoupleDoubles value = ((CoupleDoubles)it.next());
		double val = value.getSecond();
		if (val>max) max=val;
	    }
	return max;    
   }
   
   /**
    * @return String
    */
   public final String toString() 
   {
	String str = new String();

	Iterator it = keySet().iterator();
	while (it.hasNext())
	    {
		Perception p = (Perception)it.next();
		CoupleDoubles value = getCouple(p);
		str = str + p.toString() + " => " + value +"\n";
	    }
	return str;    
   }
}
