
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package MACS;

import java.util.*;
import java.lang.reflect.Array;
import Utils.StaticRandom;
import Classifier.Perception;
import Classifier.Anticipation;
import Classifier.BasicClassifier;
import Classifier.AnticipatoryClassifier;

/**
 * This class represents a list of Perceptions.
* @author <a href="mailto:Olivier.Sigaud@lip6.fr"> Olivier Sigaud</a> and <a href="mailto:gerpy@free.fr"> Pierre Grard</a>.
* 
 */
class PerceptionList extends ArrayList 
{
   
   /**
    * Creates a PerceptionList from a PerceptionSet
    * @param p the original PerceptionSet
    */
   protected PerceptionList(Set p) 
   {
	Iterator it = p.iterator();
	while  (it.hasNext())
	    add(new Perception((Perception)it.next()));    
   }
   
   /**
    * @param form
    * @param value
    */
   protected PerceptionList(String form, String value) 
   {
	StringTokenizer stringTokenizer = new StringTokenizer (value, "\n");
	do
	    {
		String perceptionValue = stringTokenizer.nextToken();
		add(new Perception(perceptionValue,form));
	    }
	while (stringTokenizer.hasMoreElements());    
   }
   
   /**
    * @param source
    */
   protected PerceptionList(PerceptionList source) 
   {
	Iterator it = source.iterator();
	while  (it.hasNext())
	    add(new Perception((Perception)it.next()));    
   }
   
   protected PerceptionList() 
   {
    
   }
   
   /**
    * Necessary because contains does not seem to work for something else than 
    * addresses
    * protected final boolean contains(Perception perc)
    * {
    * Iterator it = iterator();
    * while (it.hasNext())
    * {
    * if (((Perception)it.next()).equals(perc))return true;
    * }
    * return false;
    * }
    * 
    * @param ps
    * @return boolean
    */
   protected final boolean differsFrom(PerceptionList ps) 
   {
	Iterator it = iterator();
	Iterator it2 = ps.iterator();
	while (it.hasNext())
	    {
	    if (!((Perception)it.next()).equals(((Perception)it2.next())))	return true;
	    }
	if (it2.hasNext()) return true;
	return false;    
   }
   
   /**
    * @return Classifier.Perception
    */
   protected final Perception getRandomPerception() 
   {
	int index = (int)(StaticRandom.nextDouble()*size());
	return getPerception(index);    
   }
   
   /**
    * @return int
    */
   protected final int getNbPerceptions() 
   {
	return size();    
   }
   
   /**
    * @param index
    * @return Classifier.Perception
    */
   protected final Perception getPerception(int index) 
   {
	Iterator it = iterator();
	for (short i=0;i<index;i++) it.next();
	return (Perception)it.next();    
   }
   
   /**
    * @param perc
    * @return boolean
    */
   protected final boolean containsMatching(Perception perc) 
   {
	Iterator it = iterator();
	while (it.hasNext())
	    {
		if (((Perception)it.next()).matches(perc))	return true;
	    }
	return false;    
   }
   
   /**
    * @param c1
    * @param c2
    * @return boolean
    */
   protected final boolean matchOneIdenticalPerception(BasicClassifier c1, BasicClassifier c2) 
   {
	Iterator it = iterator();
	while (it.hasNext())
	    {
		Perception p = (Perception)it.next();
		if (c1.matchCondition(p) && c2.matchCondition(p)) return  true;
	    }
	return false;    
   }
   
   /**
    * This method merges two lists, making sure that identical elements
    * are discarded.
    * @param perceptionList the list to be merged with the original
    */
   protected final void mergeWith(PerceptionList perceptionList) 
   {
	Iterator it = perceptionList.iterator();
	while (it.hasNext())
	    {
		Perception p = (Perception)it.next();
		if (!contains(p)) add(p);
	    }
   }
   
   /**
    * From one condition part, this method produces a list of more general condition parts,
    * each member in the list generalizing over one specific attribute of the initial condition part.
    * Used by the ConditionCoveringManager and the method below.
    * @param condition a tentative condition part that should be generalized further
    * @return PerceptionList the list of more general condition parts
    */
   protected static final PerceptionList buildMoreGeneralPerceptionList(Perception condition) 
   {
	PerceptionList perceptionList = new PerceptionList();
	
	for(byte i=0 ; i<condition.getLength() ; i++)
	    if (condition.isSpecialized(i))
		{
		    Perception newPerception = new Perception(condition);
		    newPerception.generalize(i);
		    perceptionList.add(newPerception);
		}
	return perceptionList;    
   }
   
   /**
    * From one list of condition parts, this method produces a list of more general condition parts,
    * each member in the list generalizing over one specific attribute of one of the initial condition part.
    * Identical members are discarded.
    * Used by the ConditionCoveringManager.
    * @return PerceptionList
    */
   protected final PerceptionList buildMoreGeneralPerceptionList() 
   {
	PerceptionSet generalSet = new PerceptionSet();
	
	Iterator it = iterator();
	while (it.hasNext())
	    {
		generalSet.mergeWith(buildMoreGeneralPerceptionList((Perception)it.next()));
	    }
	return new PerceptionList(generalSet);    
   }
   
   /**
    * This method checks whether a classifier covers at least one Perception covered 
    * by the incompatibleClassifierList.
    * Used by the GeneralizationManager.
    * @param incompatibleClassifierList
    * @param classifier
    * @return boolean
    * @see GeneralizationManager
    */
   protected boolean coversAtLeastOnePerceptionCoveredBy(AnticipatoryClassifierList incompatibleClassifierList, AnticipatoryClassifier classifier) 
   {
	Iterator it = iterator();
	while (it.hasNext())
	    {
		Perception perception = (Perception)it.next();
		if (classifier.matchCondition(perception) 
		    && incompatibleClassifierList.containsClassifierMatching(perception))
		    return true;
	    }
	return false;    
   }
   
   /**
    * @param strPerc
    * @return int
    */
   private final int getIndexOfPerception(String strPerc) 
   {
	Iterator it = iterator();
	int cpt=0;
	while (it.hasNext())
	    {
	    if (((Perception)it.next()).toString().equals(strPerc)) return cpt;
	    cpt++;
	    }
	return -1;    
   }
   
   /**
    * This method is used by the TraceManager to organize the display.
    * 
    * @param orderSort
    * @return PerceptionList
    */
   public final PerceptionList sort(String orderSort) 
   {
	if (size()==0)    return this;
	else
	    {
		PerceptionList sortedList = new PerceptionList();
		StringTokenizer stringTokenizer = new StringTokenizer(orderSort, "\n");
		do
		    {
			int index = getIndexOfPerception(stringTokenizer.nextToken());
			if (index!=-1) sortedList.add(get(index));
		    }
		while (stringTokenizer.hasMoreElements());

		return sortedList;
	    }    
   }
   
   /**
    * @param orderSort
    * @param coords
    * @return String
    */
   public final String toString(String orderSort, String coords) 
   {
	String str = new String();
	PerceptionList sortedList =  sort(orderSort);
	Iterator it = sortedList.iterator();
	StringTokenizer stringTokenizer = new StringTokenizer (coords, "\n");
	while (stringTokenizer.hasMoreElements())
	    {
		str = str + stringTokenizer.nextToken() + "\t";
		if (it.hasNext())
		    {
			Perception p = (Perception)it.next();
			str = str + p.toString() + "\n";
		    }
		else   str=str+"---"+"\n";
	    }
	return str;    
   }
   
   /**
    * @param orderSort
    * @return String
    */
   public final String toString(String orderSort) 
   {
	return sort(orderSort).toString();    
   }
   
   /**
    * @return String
    */
   public String toString() 
   {
	String str = "[";

	Iterator it = iterator();
	while(it.hasNext())
	    str = str + ((Perception)it.next()).toString() + "\n";
	return str + "]";    
   }
}
