
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package Utils;

import	java.io.*;
import	java.util.*;

/**
 * Class that defines a reader object for parameters files.
 * <P>
 * @author <a href="http://animatlab.lip6.fr/Sigaud">Olivier Sigaud</a>
 * @version $Revision: 1.1.1.1 $
 */

public class FilesWriter implements Cloneable
{

    /**
    * Current tabulation index of the FilesWriter
    */
    protected int myTabIndex = 0;
    /**
    * Set the Current tabulation index of the FilesWriter
    */
    private final void setTabIndex (int tabIndex )
    {
        myTabIndex = tabIndex ;
    }
    /**
    * Get the Current tabulation index of the FilesWriter
    */
    private final int getTabIndex ()
    {
        return myTabIndex ;
    }
    /**
    * Add an indentation to the current tabulation index of the FilesWriter
    */
    public final void addIndent ()
    {
        myTabIndex++;
    }
    /**
    * Sub an indentation to the current tabulation index of the FilesWriter
    */
    public final void subIndent ()
    {
        myTabIndex--;
    }

    /**
     * FileOutputStream used by the FilesWriter
     */
    protected FileOutputStream myReportFileStream = null;
    /**
     * Set the FileOutputStream used by the FilesWriter
     */  
    private final void setReportFileStream(FileOutputStream myRFS) 
    {
	myReportFileStream = myRFS;
    }
    /**
     * Get the FileOutputStream used by the FilesWriter
     */  
    public final FileOutputStream getReportFileStream() 
    {
	return myReportFileStream;
    }

   /**
     * PrintWriter used by the FilesWriter
     */
    protected PrintWriter myReportWriter = null;
    /**
     * Set the PrintWriter used by the FilesWriter
     */  
    private final void setReportWriter(PrintWriter myRW) 
    {
	myReportWriter = myRW;
    }
    /**
     * Get the PrintWriter used by the FilesWriter
     */  
    public final PrintWriter getReportWriter() 
    {
	return myReportWriter;
    }

    /**
     * Name of the file used by the FilesWriter
     */
    protected String myFileName;
    /**
     * Set the Name of the file used by the FilesWriter
     */  
    private final void setFileName(String myFN) 
    {
	myFileName = myFN;
    }
    /**
     * Get the Name of the file used by the FilesWriter
     */  
    public final String getFileName() 
    {
	return myFileName;
    }

//----------------------------------------------------------------------//

    /**
     * Construct an empty FilesWriter
     */  
    public FilesWriter()
    {
    }

    /**
     * Construct the FilesWriter
     * @param myFN Name of the file used by the FilesWriter
     **/
    public FilesWriter(String myFN)
    {
	setFileName(myFN);
	open();
    }

    /**
     * Construct the FilesWriter
     * @param myDir Name of the directory used by the FilesWriter
     * @param myFN Name of the file used by the FilesWriter
     **/
    public FilesWriter(String myDir,String myFN)
    {
	boolean success = (new File(myDir)).mkdirs();
	if (!success) {
	    System.out.println("FilesWriter : mkdir failure");
	}
	setFileName(myFN);
	open();
    }

    /**
     * Open the files that the reports will be written. 
     * Called automatically by the constructor. 
     * Any re-call will re-open the files for writing, causing the previous file to be overwritten.
     **/
    public void open ()
    {
	// Try to close the files if they are already open
	close();

	// Now try to open the files
	try
	    {
		setReportFileStream(new FileOutputStream (getFileName()));
		setReportWriter(new PrintWriter (getReportFileStream(), true));
	    }
	catch (Exception e)
	    {
		// Cannot create the file, so set it to null
		setReportFileStream(null);
		e.printStackTrace();
	    }
    }

    /**
     * Close the files that the reports have been written. 
     * Automatically called when the application closes.
     **/
    public void close()
    {
	try
	    {
		if (getReportFileStream() != null)
		    getReportFileStream().close();
	    }
	catch (IOException e)
	    {
		e.printStackTrace();
		// Do nothing - will not occur if the files have opened successfully
	    }
    }

    /**
     * Print the String S in the opened file. A flush is then called
     **/
    public void outPrint(String S)
    {
	getReportWriter().print(S);
	getReportWriter().flush();
    }

    /**
     * Println the String S in the opened file. A flush is then called
     **/
    public void outPrintln(String S)
    {
	getReportWriter().println(S);
	getReportWriter().flush();
    }

    /**
     * Println the String S in the opened file. A flush is then called
     **/
    public void outTabPrintln(String S)
    {
	String s= "";
	for(int i=0; i<getTabIndex(); i++)
	    s += "  ";
	getReportWriter().println(s +S);
	getReportWriter().flush();
    }

    /**
     * Println the String S in the opened file. A flush is then called
     **/
    public void outTabPrint(String S)
    {
	String s= "";
	for(int i=0; i<getTabIndex(); i++)
	    s += "  ";
	getReportWriter().print(s +S);
	getReportWriter().flush();
    }

    /**
     * Flushes the PrintWriter
     **/
    public void flush()
    {
	getReportWriter().flush();
    }

    /**
     * Return the PrintWriter if the FilesWriter
     **/
    public PrintWriter out()
    {
	return getReportWriter();
    }

    /**
     * A test routine
     **/
    public static void main(String args[])
    {
	String	fileDir	=	"nomDir";
	String	fileName	=	"nomFichier";
	FilesWriter	writer	=	new	FilesWriter(fileDir,fileName);
	System.out.println("Ouverture du fichier :" + fileName);
	writer.open();
	(writer.out()).println("donne  inclure");
	writer.close();
    }
}
