
/**
* MACS is an Anticipatory Learning Classifier System written in Java.
* 
* Copyright (C) Pierre Grard and Olivier Sigaud with the occasional help
* of Fabien Flacher.
* Contact Olivier.Sigaud@lip6.fr for any queries, or write to
* 
* Olivier Sigaud
* AnimatLab
* Laboratoire d'Informatique de Paris 6
* 8, rue du Capitaine Scott,
* 75015 PARIS, FRANCE
* 
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
* More informations about the GNU General Public License can be found
* at the following web site:
* http://www.opensource.org/licenses/gpl-license.php
**/

package	Utils;

import java.lang.System;
import java.io.StreamTokenizer;

/**
 * Routines for units conversion. 
 * <P>
 * This class comes from Tucker Balch:
 * (c)1997, 1998 Tucker Balch
 *
 * @author Tucker Balch
 * @version $Revision: 1.4 $
 */

public class Units 
{

  /**
   * Handy to have 2 PI around.
   */
public static final double PI = Math.PI;
public static final double PI2 = 2*Math.PI;


  /**
   *  Normalize an angle into the range [0,2*PI]
   */
public static double NormalizePI(double t)
  {
    while(t>PI2) t -= PI2;
    while(t<0) t += PI2;
    return t;
  }

  /**
   *  Normalize an angle into the range [-PI,PI]
   */
public static double NormalizeZero(double t)
  {
    while(t>PI) t -= PI2;
    while(t<-PI) t += PI2;
    return t;
  }
  /**
   * A very large number
   */
public static final double HUGE = 2147483647;

  /**
   * Convert degrees to radians.
   * @param  deg double, degrees.
   * @return radians.
   */
public static double degToRad(double deg)
  {
    return(Math.PI*2*deg/360);
  }


  /**
   * Convert degrees to radians.
   * @param deg int, degrees.
   * @return radians.
   */
public static double degToRad(int deg)
  {
    return(Math.PI*2*((double)deg)/360);
  }


  /**
   * Convert 10ths of degrees to radians.
   * @param deg10 int, 10ths of degrees.
   * @return radians.
   */
public static double deg10ToRad(int deg10)
  {
    return(Math.PI*2*deg10/3600);
  }


  /**
   * Convert radians to degrees.
   * @param rad double, radians.
   * @return degrees.
   */
public static double radToDeg(double rad)
  {
    return(rad*360/(Math.PI*2));
  }


  /**
   * Convert radians to 10ths of degrees.
   * @param rad double, radians.
   * @return 10ths of degrees.
   */
public static int radToDeg10(double rad)
  {
    return((int)(rad*3600/(Math.PI*2)));
  }


  /**
   * Convert inches to meters.
   * @param inch int, inches.
   * @return meters.
   */
public static double inchToMeter(int inch)
  {
    return(((double)inch)*0.0254);
  }


  /**
   * Convert inches to meters.
   * @param inch double, inches.
   * @return meters.
   */
public static double inchToMeter(double inch)
  {
    return(inch*0.0254);
  }


  /**
   * Convert 10ths of inches to meters.
   * @param inch int, inches.
   * @return 10ths of meters.
   */
public static double inch10ToMeter(int inch10)
  {
    return(((double)inch10)*0.00254);
  }


  /**
   * Convert meters to 10ths of inches.
   * @param meter double, meters.
   * @return inches.
   */
public static int meterToInch10(double meter)
  {
    return((int)(meter/0.00254));
  }


  /**
   * Convert meters to inches.
   * @param meter double, meters.
   * @return inches.
   */
public static double meterToInch(double meter)
  {
    return(meter/0.0254);
  }


  /**
   * Limit the angle to between 0 and 2 PI.
   * @param rad double, angle in radians to be clipped.
   * @return the clipped angle.
   */
public static double clipRad(double rad)
  {
    while (rad >= PI2) rad -= PI2;
    while (rad < 0)    rad += PI2;
    return(rad);
  }


  /**
   * Limit the angle to between 0 and 359.9999.
   * @param deg double, angle in degrees to be clipped.
   * @return the clipped angle.
   */
public static double clipDeg(double deg)
  {
    while (deg >= 360) deg -= 360;
    while (deg < 0)    deg += 360;
    return(deg);
  }


  /**
   * Compute the best direction and angle to turn from the start 
   * angle to the finish angle in degrees.
   * @param start The starting angle.
   * @param finish The desired angle.
   * @return the required turn.
   */
public static double bestTurnDeg(double start, double finish)
  {
    /*--- normalize to between 0 and 360 ---*/
    while (start >= 360) start -= 360;
    while (start < 0)   start += 360;
    while (finish >= 360) finish -= 360;
    while (finish < 0)   finish += 360;

    /*--- compute best turn ---*/
    double rot = finish - start;
    if (rot > 180) rot -= 360;
    else if (rot < -180) rot += 360;

    return(rot);
  }


  /**
   * read a double from a stream. Assumes you have just read the
   * token.
   *
   * @param in StreamTokenizer, the stream to read from.
   * @return the double.
   * @exception Exception if there is no number there.
   */
public static double readDouble(StreamTokenizer in) throws Exception
  {
    double num = in.nval;
    int exp = 0;
    in.ordinaryChars('\0', ' ');
    in.nextToken();
    in.whitespaceChars('\0', ' ');
    if (in.ttype == in.TT_WORD &&
	Character.toUpperCase(in.sval.charAt(0))=='E')
      {
	try
	  {
	    exp = Integer.parseInt(in.sval.substring(1));
	  }
	catch(NumberFormatException e)
	  {
	    in.pushBack();
	  }
      }
    num = num * Math.pow(10,exp);
    return(num);
  }


  /**
   * Compute the best direction and angle to turn from the start 
   * angle to the finish angle in radians.
   * @param start The starting angle.
   * @param finish The desired angle.
   * @return The required turn.
   */
public static double bestTurnRad(double start, double finish)
  {
    /*--- normalize to between 0 and 2PI ---*/
    while (start >= PI2) start -= PI2;
    while (start < 0)   start += PI2;
    while (finish >= PI2) finish -= PI2;
    while (finish < 0)   finish += PI2;

    /*--- compute best turn ---*/
    double rot = finish - start;
    if (rot > Math.PI) rot -= PI2;
    else if (rot < -Math.PI) rot += PI2;

    return(rot);
  }


  /**
   * Test the units functions.
   */
public static void main(String[] args)
  {
    if (degToRad(360.0)!=(2*Math.PI))
      System.out.println("bug1");
    else if (degToRad((int)360)!=(2*Math.PI))
      System.out.println("bug2");
    else if (deg10ToRad((int)3600)!=(2*Math.PI))
      System.out.println("bug3");
    else if (radToDeg(2*Math.PI)!=360.0)
      System.out.println("bug4");
    else if (radToDeg10(2*Math.PI)!=3600)
      System.out.println("bug5");
    else if (inch10ToMeter(400)!=1.016)
      System.out.println("bug6");
    else if (meterToInch10(1.016)!=400)
      System.out.println("bug7");
    else if (bestTurnDeg(10,20)!=10)
      System.out.println("bug8");
    else if (bestTurnDeg(1,360)!=-1)
      System.out.println("bug9 "+bestTurnDeg(1,360));
    else if (bestTurnDeg(1,359)!=-2)
      System.out.println("bug10 "+bestTurnDeg(1,359));
    else if (bestTurnDeg(359,1)!=2)
      System.out.println("bug11");
    else if (bestTurnDeg(10,200)!=-170)
      System.out.println("bug12");
    else if (bestTurnDeg(200,10)!=170)
      System.out.println("bug13");
    else if (bestTurnDeg(721,1079)!=-2)
      System.out.println("bug14");
    else /* no bugs */
      System.out.println("no bugs detected");
  }
}
